import yucca
from batchgenerators.utilities.file_and_folder_operations import join, load_json
from dataclasses import dataclass
from typing import Union, Literal
from yucca.pipeline.preprocessing.UnsupervisedPreprocessor import UnsupervisedPreprocessor
from yucca.pipeline.preprocessing.ClassificationPreprocessor import ClassificationPreprocessor
from yucca.functional.utils.dict import without_keys
from yucca.functional.utils.files_and_folders import recursive_find_python_class
import logging


@dataclass
class PlanConfig:
    allow_missing_modalities: bool
    image_extension: str
    num_classes: int
    plans: dict
    task_type: str
    use_label_regions: bool = False
    regions_in_order: list = None
    regions_labeled: list = None

    def lm_hparams(self, without: [] = []):
        hparams = {
            "allow_missing_modalities": self.allow_missing_modalities,
            "image_extension": self.image_extension,
            "num_classes": self.num_classes,
            "plans": self.plans,
            "task_type": self.task_type,
            "regions_in_order": self.regions_in_order,
            "regions_labeled": self.regions_labeled,
            "use_label_regions": self.use_label_regions,
        }
        return without_keys(hparams, without)


def get_plan_config(
    plans_path: str,
    stage: Literal["fit", "test", "predict"],
    ckpt_plans: Union[dict, None] = None,
    use_label_regions: bool = False,
):
    assert stage in ["fit", "test", "predict"], f"stage: {stage} is not supported"
    # First try to load torch checkpoints and extract plans and carry-over information from there.
    if stage == "fit":
        plans = load_plans(plans_path)
    if stage == "test":
        raise NotImplementedError
    if stage == "predict":
        # In this case we don't want to rely on plans being found in the preprocessed folder,
        # as it might not exist.
        assert ckpt_plans is not None
        plans = ckpt_plans

    regions_in_order = None
    regions_labeled = None

    task_type = setup_task_type(plans)
    if task_type == "self-supervised":
        num_classes = max(1, plans.get("num_modalities") or len(plans["dataset_properties"]["modalities"]))
    elif use_label_regions:
        regions_in_order = plans["dataset_properties"]["regions_in_order"]
        regions_labeled = plans["dataset_properties"]["regions_labeled"]
        num_classes = len(regions_in_order)
    else:
        num_classes = max(1, plans.get("num_classes") or len(plans["dataset_properties"]["classes"]))
    image_extension = plans.get("image_extension") or plans["dataset_properties"].get("image_extension") or "nii.gz"
    allow_missing_modalities = plans.get("allow_missing_modalities") or False

    return PlanConfig(
        allow_missing_modalities=allow_missing_modalities,
        image_extension=image_extension,
        num_classes=num_classes,
        plans=plans,
        task_type=task_type,
        use_label_regions=use_label_regions,
        regions_in_order=regions_in_order,
        regions_labeled=regions_labeled,
    )


def load_plans(plans_path):
    # If plans is still none the ckpt files were either empty/invalid or didn't exist and we load the plans
    # from the preprocessed folder.
    keys_to_drop = ["original_sizes", "original_spacings", "new_sizes", "new_spacings"]
    logging.info("Loading plans.json")
    plan = load_json(plans_path)
    for key in keys_to_drop:
        if key in plan:
            plan.pop(key)
        elif key in plan["dataset_properties"]:
            plan["dataset_properties"].pop(key)
    return plan


def setup_task_type(plans):
    if plans.get("task_type"):
        task_type = plans.get("task_type")
        return task_type

    # If key is not present in plan then we try to infer the task_type from the Type of Preprocessor
    preprocessor_class = recursive_find_python_class(
        folder=[join(yucca.__path__[0], "pipeline", "preprocessing")],
        class_name=plans["preprocessor"],
        current_module="yucca.pipeline.preprocessing",
    )
    assert (
        preprocessor_class
    ), f"{plans['preprocessor']} was found in plans, but no class with the corresponding name was found"
    if issubclass(preprocessor_class, ClassificationPreprocessor):
        task_type = "classification"
    elif issubclass(preprocessor_class, UnsupervisedPreprocessor):
        task_type = "self-supervised"
    else:
        task_type = "segmentation"
    return task_type



from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.pipeline.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data
from tqdm import tqdm
import nibabel as nib


def convert(path: str, subdir: str = "decathlon/Task01_BrainTumour"):
    """INPUT DATA - Define input path and suffixes"""
    path = join(path, subdir)
    ext = ".nii.gz"

    """ OUTPUT DATA - Define the task name and prefix """
    task_name = "Task206_BrainTumour"
    task_prefix = "BrainTumour"

    """ Access the input data. If images are not split into train/test, and you wish to randomly
    split the data, uncomment and adapt the following lines to fit your local path. """

    # NOTE: We use the test set for pre-training, as the labels are no longer available, and we thus cannot use it for evaluation!
    subjects_dir = join(path, "imagesTs")
    target_base = join(yucca_raw_data, task_name)
    target_imagesTr = join(target_base, "imagesTr")

    maybe_mkdir_p(target_imagesTr)

    """Populate Target Directory
    This is also the place to apply any re-orientation, resampling and/or label correction."""

    for subject in tqdm(subfiles(subjects_dir, join=False, suffix=ext)):
        image_path = join(subjects_dir, subject)
        other_info = subject[6 : -len(ext)]  # remove the prefix BRATS_ and the suffix .nii.gz

        vol = nib.load(image_path)

        flair = vol.slicer[:, :, :, 0]
        t1w = vol.slicer[:, :, :, 1]
        t1gd = vol.slicer[:, :, :, 2]
        t2w = vol.slicer[:, :, :, 3]

        nib.save(flair, filename=f"{target_imagesTr}/{task_prefix}_{subject}_flair_{other_info}_000.nii.gz")
        nib.save(t1w, filename=f"{target_imagesTr}/{task_prefix}_{subject}_t1w_{other_info}_000.nii.gz")
        nib.save(t1gd, filename=f"{target_imagesTr}/{task_prefix}_{subject}_t1gd_{other_info}_000.nii.gz")
        nib.save(t2w, filename=f"{target_imagesTr}/{task_prefix}_{subject}_t2w_{other_info}_000.nii.gz")

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        imagesTs_dir=None,
        modalities=["MRI"],
        labels=None,
        dataset_name=task_name,
        license="CC-BY-SA 4.0",
        dataset_description="Decathlon: Brain Tumour Segmentation",
        dataset_reference="King's College London",
    )



from dataclasses import dataclass
from typing import Union


@dataclass
class TaskConfig:
    continue_from_most_recent: bool
    experiment: str
    manager_name: str
    model_dimensions: str
    model_name: str
    patch_based_training: bool
    planner_name: str
    split_idx: int
    split_method: str
    split_param: Union[str, float, int]
    task: str

    def lm_hparams(self):
        return {
            "continue_from_most_recent": self.continue_from_most_recent,
            "experiment": self.experiment,
            "manager_name": self.manager_name,
            "model_dimensions": self.model_dimensions,
            "model_name": self.model_name,
            "patch_based_training": self.patch_based_training,
            "planner_name": self.planner_name,
            "task": self.task,
            "split_idx": self.split_idx,
            "split_method": self.split_method,
            "split_param": self.split_param,
        }


def get_task_config(
    task: str,
    continue_from_most_recent: bool = True,
    manager_name: str = "YuccaManager",
    model_dimensions: str = "3D",
    model_name: str = "UNet",
    planner_name: str = "YuccaPlanner",
    patch_based_training: bool = True,
    experiment: str = "default",
    split_idx: int = 0,
    split_data_method: str = "kfold",
    split_data_param: Union[str, float, int] = 5,
):
    assert model_dimensions is not None

    return TaskConfig(
        task=task,
        continue_from_most_recent=continue_from_most_recent,
        manager_name=manager_name,
        model_dimensions=model_dimensions,
        model_name=model_name,
        patch_based_training=patch_based_training,
        planner_name=planner_name,
        experiment=experiment,
        split_idx=split_idx,
        split_method=split_data_method,
        split_param=split_data_param,
    )



import torch
import torch.nn as nn
from yucca.networks.networks.YuccaNet import YuccaNet
from yucca.networks.blocks_and_layers.conv_layers import (
    DoubleConvDropoutNormNonlin,
)


class TinyUNet(YuccaNet):
    def __init__(
        self,
        input_channels: int,
        num_classes: int = 1,
        starting_filters: int = 4,
        conv_op=nn.Conv2d,
        conv_kwargs={
            "kernel_size": 3,
            "stride": 1,
            "padding": 1,
            "dilation": 1,
            "bias": True,
        },
        norm_op=nn.InstanceNorm2d,
        norm_op_kwargs={"eps": 1e-5, "affine": True, "momentum": 0.1},
        dropout_op=nn.Dropout2d,
        dropout_op_kwargs={"p": 0.0, "inplace": True},
        nonlin=nn.LeakyReLU,
        nonlin_kwargs={"negative_slope": 1e-2, "inplace": True},
        dropout_in_decoder=False,
        weightInitializer=None,
        basic_block=DoubleConvDropoutNormNonlin,
        deep_supervision=False,
    ) -> None:
        super(TinyUNet, self).__init__()

        # Task specific
        self.num_classes = num_classes
        self.filters = starting_filters

        # Model parameters
        self.conv_op = conv_op
        self.conv_kwargs = conv_kwargs
        self.norm_op_kwargs = norm_op_kwargs
        self.norm_op = norm_op
        self.dropout_op = dropout_op
        self.dropout_op_kwargs = dropout_op_kwargs
        self.nonlin_kwargs = nonlin_kwargs
        self.nonlin = nonlin
        self.weightInitializer = weightInitializer
        self.basic_block = basic_block
        self.deep_supervision = deep_supervision

        # Dimensions
        if self.conv_op == nn.Conv2d:
            self.pool_op = torch.nn.MaxPool2d
            self.upsample = torch.nn.ConvTranspose2d
        else:
            self.pool_op = torch.nn.MaxPool3d
            self.upsample = torch.nn.ConvTranspose3d

        self.in_conv = self.basic_block(
            input_channels,
            self.filters,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
        )

        self.pool1 = self.pool_op(2)
        self.encoder_conv1 = self.basic_block(
            self.filters,
            self.filters * 2,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
        )

        self.pool2 = self.pool_op(2)
        self.encoder_conv2 = self.basic_block(
            self.filters * 2,
            self.filters * 4,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
        )

        # Decoder
        if not dropout_in_decoder:
            old_dropout_p = self.dropout_op_kwargs["p"]
            self.dropout_op_kwargs["p"] = 0.0

        self.upsample1 = self.upsample(self.filters * 4, self.filters * 2, kernel_size=2, stride=2)
        self.decoder_conv1 = self.basic_block(
            self.filters * 4,
            self.filters * 2,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
        )

        self.upsample2 = self.upsample(self.filters * 2, self.filters, kernel_size=2, stride=2)
        self.decoder_conv2 = self.basic_block(
            self.filters * 2,
            self.filters,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
        )

        self.out_conv = self.conv_op(self.filters, self.num_classes, kernel_size=1)

        if not dropout_in_decoder:
            self.dropout_op_kwargs["p"] = old_dropout_p

        if self.weightInitializer is not None:
            print("initializing weights")
            self.apply(self.weightInitializer)

    def forward(self, x):
        x0 = self.in_conv(x)

        # Encoder path
        x1 = self.pool1(x0)
        x1 = self.encoder_conv1(x1)

        x2 = self.pool2(x1)
        x2 = self.encoder_conv2(x2)

        # Decoder path
        x3 = torch.cat([self.upsample1(x2), x1], dim=1)
        x3 = self.decoder_conv1(x3)

        x4 = torch.cat([self.upsample2(x3), x0], dim=1)
        x4 = self.decoder_conv2(x4)

        if self.deep_supervision:
            pass

        logits = self.out_conv(x4)
        return logits



import numpy as np
import torch
import nibabel as nib
import os
import logging
import time
import re
from yucca.functional.testing.data.nifti import (
    verify_spacing_is_equal,
    verify_orientation_is_equal,
)
from yucca.functional.testing.data.array import verify_labels_are_equal, verify_array_shape_is_equal
from yucca.functional.preprocessing import (
    preprocess_case_for_inference,
    preprocess_case_for_training_with_label,
    preprocess_case_for_training_without_label,
    reverse_preprocessing,
)
from yucca.functional.utils.loading import load_yaml, read_file_to_nifti_or_np
from yucca.paths import yucca_preprocessed_data, yucca_raw_data
from multiprocessing import Pool
from batchgenerators.utilities.file_and_folder_operations import (
    join,
    load_json,
    subfiles,
    save_pickle,
    maybe_mkdir_p,
    isfile,
)


class YuccaPreprocessor(object):
    """
    The YuccaPreprocessor class is designed to preprocess medical images for the Yucca project.
    It implements various preprocessing steps, such as reorientation, cropping, normalization, and resizing,
    based on the plans specified in an YuccaPlanner.

    For training the _preprocess_train_subject method prepares input images for the Yucca model.
    The preprocess_case_for_inference method prepares input images for the Yucca model during the inference phase,
    ensuring that they match the requirements specified during training.
    The reverse_preprocessing method is then used to revert the processed images back to their original form,
    allowing for a meaningful interpretation of the model's predictions.
    These methods collectively provide a consistent and reversible preprocessing pipeline for both training and inference.

    The operations that can be enabled/defined in the YuccaPlanner and carried out by the
    YuccaPreprocessor are:

    (1) The starting orientation - defaults to RAS (for medical images).
    (2) The cropping operation - defaults to crop to nonzero bounding box
    (3) The Transposition operation (along with the reverse transpose operation,
    to be used during inference) - defaults to no transposition if image dimensions and spacings
    are not too anisotropic.
    (4) The Resample operation - defaults to resampling to the median spacing of the dataset.
    (5) The Normalization operation - defaults to standardization = (image - mean) / std
    per modality to preserve ranges to account for CT pixel values representing specific physical
    attributes.

    Additionally it carries out a number of tests and analyzes each image for foreground locations
    which is used later to oversample foreground.
    """

    def __init__(
        self,
        plans_path,
        task=None,
        threads=None,
        disable_sanity_checks=False,
        enable_cc_analysis=False,
        allow_missing_modalities=False,
        compress=False,
        get_foreground_locs_per_label=False,
    ):
        self.name = str(self.__class__.__name__)
        self.task = task
        self.plans_path = plans_path
        self.plans = self.load_plans(plans_path)
        self.threads = threads
        self.disable_sanity_checks = disable_sanity_checks
        self.enable_cc_analysis = enable_cc_analysis
        self.allow_missing_modalities = allow_missing_modalities
        self.compress = compress
        self.get_foreground_locs_per_label = get_foreground_locs_per_label

        # lists for information we would like to attain
        self.transpose_forward = []
        self.transpose_backward = []
        self.target_spacing = []

        # set up for segmentation
        self.classification = False
        self.label_exists = True
        self.preprocess_label = True

    def initialize_paths(self):
        self.target_dir = join(yucca_preprocessed_data, self.task, self.plans["plans_name"])
        self.input_dir = join(yucca_raw_data, self.task)
        self.imagepaths = subfiles(join(self.input_dir, "imagesTr"), suffix=self.image_extension)
        self.subject_ids = [
            file for file in subfiles(join(self.input_dir, "labelsTr"), join=False) if not file.startswith(".")
        ]

    def initialize_properties(self):
        """
        here we basically set up things that are needed for preprocessing during training,
        but that aren't necessary during inference
        """
        self.dataset_properties = self.plans["dataset_properties"]
        self.intensities = self.dataset_properties["intensities"]
        self.image_extension = self.dataset_properties.get("image_extension") or "nii.gz"

        if self.dataset_properties.get("background_pixel_values") is not None:
            self.background_value_for_first_modality = self.dataset_properties.get("background_pixel_values")[0]
        else:
            self.background_value_for_first_modality = 0

        # op values
        self.transpose_forward = np.array(self.plans["transpose_forward"], dtype=int)
        self.transpose_backward = np.array(self.plans["transpose_backward"], dtype=int)
        self.target_spacing = self.plans["target_spacing"]
        self.target_size = (
            np.array(self.plans.get("target_size"), dtype=int) if self.plans.get("target_size") not in ["null", None] else None
        )

    def run(self):
        self.initialize_properties()
        self.initialize_paths()
        maybe_mkdir_p(self.target_dir)
        self.verify_compression_level(self.target_dir, self.compress)

        logging.info(
            f"{'Preprocessing Task:':25.25} {self.task} \n"
            f"{'Using Planner:':25.25} {self.plans_path} \n"
            f"{'Crop to nonzero:':25.25} {self.plans['crop_to_nonzero']} \n"
            f"{'Normalization scheme:':25.25} {self.plans['normalization_scheme']} \n"
            f"{'Transpose Forward:':25.25} {self.transpose_forward} \n"
            f"{'Transpose Backward:':25.25} {self.transpose_backward} \n"
            f"{'Number of threads:':25.25} {self.threads}"
        )
        p = Pool(self.threads)
        p.map(self.preprocess_train_subject, self.subject_ids)
        p.close()
        p.join()

    def preprocess_train_subject(self, subject_id):
        """
        This is the bread and butter of the preprocessor.
        The following steps are taken:

        (1) Load Images:
        Extract relevant image files associated with the given subject_id.
        Load the images using the nibabel library.

        (2) Reorientation (Optional):
        Check if valid qform or sform codes are present in the header.
        If valid, reorient the images to the target orientation specified in the plans.
        Update the original and new orientation information in the image_props dictionary.

        (3) Normalization and Transposition:
        Normalize each image based on the specified normalization scheme and intensities.
        Transpose the images according to the forward transpose axes specified in the plans.

        (4) Cropping (Optional):
        If the crop_to_nonzero option is enabled in the plans, crop the images to the nonzero bounding box.
        Update the image_props dictionary with cropping information.

        (5) Resampling:
        Resample images to the target spacing specified in the plans.
        Update the image_props dictionary with original and new spacing information.

        (6) Foreground Locations:
        Extract some locations of the foreground, which will be used in oversampling of foreground classes.
        Determine the number and sizes of connected components in the ground truth label (can be used in analysis).

        (7) Save Preprocessed Data:
        Stack the preprocessed images and label.
        Save the preprocessed data as a NumPy array in a .npy file.
        Save relevant metadata as a .pkl file.

        (8) Print Information:
        Print information about the size and spacing before and after preprocessing.
        Print the path where the preprocessed data is saved.
        """
        subject_id = subject_id.split(os.extsep, 1)[0]
        if self.compress:
            arraypath = join(self.target_dir, subject_id + ".npz")
        else:
            arraypath = join(self.target_dir, subject_id + ".npy")
        picklepath = join(self.target_dir, subject_id + ".pkl")
        if isfile(arraypath) and isfile(picklepath):
            logging.info(f"Case: {subject_id} already exists. Skipping.")
            return

        start_time = time.time()

        images, label, image_props = self._preprocess_train_subject(
            subject_id, label_exists=self.label_exists, preprocess_label=self.preprocess_label
        )
        images = self.cast_to_numpy_array(images=images, label=label, classification=self.classification)

        # save the image
        if self.compress:
            np.savez_compressed(arraypath, data=images)
        else:
            np.save(arraypath, images)

        # save metadata as .pkl
        save_pickle(image_props, picklepath)

        end_time = time.time()
        logging.info(
            f"Preprocessed case: {subject_id} \n"
            f"size before: {image_props['original_size']} \n"
            f"cropping enabled: {self.plans['crop_to_nonzero']} \n"
            f"size after crop: {image_props['size_before_transpose']} \n"
            f"size final: {image_props['new_size']} \n"
            f"spacing before: {image_props['original_spacing']} spacing after: {image_props['new_spacing']} \n"
            f"Saving {subject_id} in {arraypath} \n"
            f"Time elapsed: {round(end_time-start_time, 4)} \n"
        )
        del images, label, image_props

    def _preprocess_train_subject(self, subject_id, label_exists: bool, preprocess_label: bool):
        image_props = {}

        # First find relevant images by their paths and save them in the image property pickle
        # Then load them as images
        # The '_' in the end is to avoid treating Case_4_000 AND Case_42_000 as different versions
        # of the label named Case_4 as both would start with "Case_4", however only the correct one is
        # followed by an underscore
        escaped_subject_id = re.escape(subject_id)
        # path to all modalities of subject_id
        imagepaths = [
            impath
            for impath in self.imagepaths
            # Check if impath is a modality of subject_id (subject_id + _XXX + .) where XXX are three digits
            if re.search(escaped_subject_id + "_" + r"\d{3}" + ".", os.path.split(impath)[-1])
        ]

        expected_modalities = set([f"{i:03}" for i in range(len(self.plans["normalization_scheme"]))])
        found_modalities = [os.path.split(impath)[-1].split(os.extsep, 1)[0][-3:] for impath in imagepaths]
        missing_modalities = [int(missing_mod) for missing_mod in list(expected_modalities.difference(found_modalities))]

        assert len(imagepaths) > 0, "found no images"
        if not self.allow_missing_modalities:
            assert not len(missing_modalities) > 0, "found missing modalities and allow_missing_modalities is not enabled."
        image_props["image files"] = imagepaths
        images = [read_file_to_nifti_or_np(image) for image in imagepaths]

        if label_exists:
            # Do the same with label
            label = [
                labelpath
                for labelpath in subfiles(join(self.input_dir, "labelsTr"))
                if os.path.split(labelpath)[-1].startswith(subject_id + ".")
            ]
            assert len(label) == 1, f"unexpected number of labels found. Expected 1 and found {len(label)}"
            image_props["label file"] = label[0]
            label = read_file_to_nifti_or_np(label[0], dtype=np.uint8)
        else:
            label = None

        if not self.disable_sanity_checks:
            if label_exists and preprocess_label:
                self.run_sanity_checks(images, label, subject_id, imagepaths)
            else:
                self.run_sanity_checks(images, None, subject_id, imagepaths)

        if label_exists and preprocess_label:
            images, label, image_props = preprocess_case_for_training_with_label(
                images=images,
                label=label,
                normalization_operation=self.plans["normalization_scheme"],
                allow_missing_modalities=self.allow_missing_modalities,
                background_pixel_value=self.background_value_for_first_modality,
                enable_cc_analysis=self.enable_cc_analysis,
                foreground_locs_per_label=self.get_foreground_locs_per_label,
                missing_modality_idxs=missing_modalities,
                crop_to_nonzero=self.plans["crop_to_nonzero"],
                keep_aspect_ratio_when_using_target_size=self.plans["keep_aspect_ratio_when_using_target_size"],
                image_properties=image_props,
                intensities=self.intensities,
                target_orientation=self.plans["target_coordinate_system"],
                target_size=self.target_size,
                target_spacing=self.target_spacing,
                transpose=self.transpose_forward,
            )
            self.verify_label_validity(label, subject_id)

        else:
            images, image_props = preprocess_case_for_training_without_label(
                images=images,
                normalization_operation=self.plans["normalization_scheme"],
                allow_missing_modalities=self.allow_missing_modalities,
                background_pixel_value=self.background_value_for_first_modality,
                missing_modality_idxs=missing_modalities,
                crop_to_nonzero=self.plans["crop_to_nonzero"],
                keep_aspect_ratio_when_using_target_size=self.plans["keep_aspect_ratio_when_using_target_size"],
                image_properties=image_props,
                intensities=self.intensities,
                target_orientation=self.plans["target_coordinate_system"],
                target_size=self.target_size,
                target_spacing=self.target_spacing,
                transpose=self.transpose_forward,
            )
        return images, label, image_props

    def preprocess_case_for_inference(self, images: list | tuple, patch_size: tuple, sliding_window_prediction: bool = True):
        """
        Will reorient ONLY if we have valid qform or sform codes.
        with coded=True the methods will return {affine or None} and {0 or 1}.
        If both are 0 we cannot rely on headers for orientations and will
        instead assume images are in the desired orientation already.

        Afterwards images will be normalized and transposed as specified by the
        plans file also used in training.

        Finally images are resampled to the required spacing/size and returned
        as torch tensors of the required shape (b, c, x, y, (z))
        """
        assert isinstance(images, (list, tuple)), "image(s) should be a list or tuple, even if only one " "image is passed"
        self.initialize_properties()

        if sliding_window_prediction is False:
            self.target_size = patch_size

        images, image_properties = preprocess_case_for_inference(
            crop_to_nonzero=self.plans["crop_to_nonzero"],
            keep_aspect_ratio=self.plans["keep_aspect_ratio_when_using_target_size"],
            images=images,
            intensities=self.intensities,
            normalization_scheme=self.plans["normalization_scheme"],
            patch_size=patch_size,
            target_size=self.target_size,
            target_spacing=self.target_spacing,
            target_orientation=self.plans["target_coordinate_system"],
            transpose_forward=self.transpose_forward,
            allow_missing_modalities=self.allow_missing_modalities,
        )

        return images, image_properties

    def reverse_preprocessing(self, images: torch.Tensor, image_properties: dict):
        """
        Expected shape of images are:
        (b, c, x, y(, z))

        (1) Initialization: Extract relevant properties from the image_properties dictionary.
        (2) Padding Reversion: Reverse the padding applied during preprocessing.
        (3) Resampling and Transposition Reversion: Resize the images to revert the resampling operation.
        Transpose the images back to the original orientation.
        (4) Cropping Reversion (Optional): If cropping to the nonzero bounding box was applied, revert the cropping operation.
        (5) Return: Return the reverted images as a NumPy array.
        The original orientation of the image will be re-applied when saving the prediction
        """
        nclasses = max(1, len(self.plans["dataset_properties"]["classes"]))

        images, image_properties = reverse_preprocessing(
            crop_to_nonzero=self.plans["crop_to_nonzero"],
            images=images,
            image_properties=image_properties,
            n_classes=nclasses,
            transpose_forward=self.transpose_forward,
            transpose_backward=self.transpose_backward,
        )

        return images, image_properties

    @staticmethod
    def load_plans(plans_path):
        if os.path.splitext(plans_path)[-1] == ".json":
            return load_json(plans_path)
        if os.path.splitext(plans_path)[-1] == ".yaml":
            return load_yaml(plans_path)["config"]["plans"]
        else:
            raise FileNotFoundError(
                f"Plan file not found. Got {plans_path} with ext {os.path.splitext(plans_path)[-1]}. Expects either a '.json' or '.yaml' file."
            )

    def run_sanity_checks(self, images, label, subject_id, imagepaths):
        self.sanity_check_standard_formats(images, label, subject_id, imagepaths)

        if isinstance(images[0], nib.Nifti1Image):
            self.sanity_check_niftis(images, label, subject_id)

    def sanity_check_standard_formats(self, images, label, subject_id, imagepaths):
        assert len(images) > 0, f"found no images for {subject_id + '_'}, " f"attempted imagepaths: {imagepaths}"

        assert (
            len(images[0].shape) == self.plans["dataset_properties"]["data_dimensions"]
        ), f"image should be shape (x, y(, z)) but is {images[0].shape}"

        if label is not None:
            verify_array_shape_is_equal(reference=images[0], target=label, id=subject_id)

        # Make sure all modalities are correctly registered
        if len(images) > 1:
            for image in images:
                verify_array_shape_is_equal(reference=images[0], target=image, id=subject_id)

    def sanity_check_niftis(self, images, label, subject_id):
        if label is not None:
            verify_spacing_is_equal(reference=images[0], target=label, id=subject_id)
            verify_orientation_is_equal(reference=images[0], target=label, id=subject_id)

        if len(images) > 1:
            for image in images:
                verify_spacing_is_equal(reference=images[0], target=image, id=subject_id)
                verify_orientation_is_equal(reference=images[0], target=image, id=subject_id)

    def cast_to_numpy_array(self, images: list, label=None, classification=False):
        if label is None and not self.allow_missing_modalities:  # self-supervised
            images = np.array(images, dtype=np.float32)
        elif label is None and self.allow_missing_modalities:  # self-supervised with missing mods
            images = np.array(images, dtype="object")
        elif classification:  # Classification is always "object"
            images.append(label)
            images = np.array(images, dtype="object")
        elif self.allow_missing_modalities:  # segmentation with missing modalities
            images.append(np.array(label)[np.newaxis])
            images = np.array(images, dtype="object")
        else:  # Standard segmentation
            images = np.vstack((np.array(images), np.array(label)[np.newaxis]), dtype=np.float32)
        return images

    def verify_label_validity(self, label, subject_id):
        # Check if the ground truth only contains expected values
        expected_labels = np.array(self.plans["dataset_properties"]["classes"], dtype=np.float32)
        actual_labels = np.unique(label).astype(np.float32)
        verify_labels_are_equal(expected_labels=expected_labels, actual_labels=actual_labels, id=subject_id)

    @staticmethod
    def verify_compression_level(directory: str, compress: bool):
        if compress:
            assert (
                len(subfiles(directory, suffix=".npy")) == 0
            ), "detected uncompressed images in folder while compression is enabled. Please delete uncompressed images first to avoid duplicates"
        else:
            assert (
                len(subfiles(directory, suffix=".npz")) == 0
            ), "detected compressed images in folder while compression is disabled. Please delete compressed images first to avoid duplicates"



import shutil
import nibabel as nib
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles, subdirs
from yucca.pipeline.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data, yucca_source


def convert(path: str = yucca_source, subdir: str = "ATLAS_2"):
    """INPUT DATA - Define input path and suffixes"""
    path = join(path, subdir)
    file_suffix = ".nii.gz"

    """ OUTPUT DATA - Define the task name and prefix """
    task_name = "Task020_ATLAS2"
    task_prefix = "ATLAS2"

    """ Access the input data. If images are not split into train/test, and you wish to randomly 
    split the data, uncomment and adapt the following lines to fit your local path. """

    images_dir_tr = labels_dir_tr = join(path, "Training")
    # images_dir_ts = labels_dir_ts = join(path, "Testing")

    """ Then define target paths """
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    """Populate Target Directory
    This is also the place to apply any re-orientation, resampling and/or label correction."""

    for dataset in subdirs(images_dir_tr, join=False):
        for sTr in subdirs(join(images_dir_tr, dataset), join=False):
            sessions = subdirs(join(images_dir_tr, dataset, sTr), join=False)
            assert len(sessions) == 1 and sessions[0] == "ses-1", f"unexpected # sessions found: {sessions}"
            image_types = subdirs(join(images_dir_tr, dataset, sTr, sessions[0]), join=False)
            assert len(image_types) == 1 and image_types[0] == "anat", f"unexpected type of scan. Found {image_types}"

            image_and_label = subfiles(join(images_dir_tr, dataset, sTr, sessions[0], image_types[0]), join=False)
            image = [i for i in image_and_label if "label" not in i][0]
            label = [i for i in image_and_label if "label" in i][0]
            case_id = image[: -len(file_suffix)]

            src_image_file_path = join(images_dir_tr, dataset, sTr, sessions[0], image_types[0], image)
            src_label_file_path = join(labels_dir_tr, dataset, sTr, sessions[0], image_types[0], label)

            label = nib.load(src_label_file_path)
            labelarr = label.get_fdata()
            labelarr[labelarr > 0] = 1
            label = nib.Nifti1Image(labelarr, label.affine, label.header)

            dst_image_file_path = f"{target_imagesTr}/{task_prefix}_{case_id}_000.nii.gz"
            dst_label_file_path = f"{target_labelsTr}/{task_prefix}_{case_id}.nii.gz"

            shutil.copy2(src_image_file_path, dst_image_file_path)
            nib.save(label, dst_label_file_path)

    # for sTs in subfiles(images_dir_ts, suffix=file_suffix, join=False):
    #    case_id = sTs[: -len(file_suffix)]
    #    src_image_file_path = join(images_dir_ts, case_id + file_suffix)
    #    dst_image_file_path = f"{target_imagesTs}/{task_prefix}_{case_id}_000.nii.gz"
    #    shutil.copy2(src_image_file_path, dst_image_file_path)

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        modalities=("T1",),
        labels={
            0: "background",
            1: "lesion",
        },
        dataset_name=task_name,
        license="",
        dataset_description="ATLAS2 from MICCAI 2022",
        dataset_reference="https://atlas.grand-challenge.org/ATLAS/",
    )


if __name__ == "__main__":
    convert()



import lightning as L
import torch
import wandb
import logging
from typing import Literal, Union, Optional
from yucca.data.augmentation.YuccaAugmentationComposer import YuccaAugmentationComposer
from yucca.pipeline.configuration.split_data import get_split_config, SplitConfig
from yucca.pipeline.configuration.configure_task import get_task_config
from yucca.pipeline.configuration.configure_callbacks import get_callback_config
from yucca.pipeline.configuration.configure_checkpoint import get_checkpoint_config
from yucca.pipeline.configuration.configure_seed import seed_everything_and_get_seed_config
from yucca.pipeline.configuration.configure_paths import get_path_config
from yucca.pipeline.configuration.configure_plans import get_plan_config
from yucca.pipeline.configuration.configure_input_dims import get_input_dims_config
from yucca.data.data_modules.YuccaDataModule import YuccaDataModule
from yucca.data.datasets.YuccaDataset import YuccaTrainDataset, YuccaTestDataset
from yucca.data.samplers import InfiniteRandomSampler
from yucca.lightning_modules.YuccaLightningModule import YuccaLightningModule
from yucca.paths import yucca_results
from yucca.functional.utils.torch_utils import measure_FLOPs
from fvcore.nn import flop_count_table


class YuccaManager:
    """
    The YuccaLightningManager class provides a convenient way to manage the training and inference processes in the Yucca project.
    It encapsulates the configuration, setup, and execution steps, making it easier to conduct experiments and predictions with consistent settings.

    The initialize method of the YuccaLightningManager class is responsible for setting up the necessary components for either training or inference.
    This method performs the configuration of paths, creates data augmentation objects, and sets up the PyTorch Lightning modules (model and data module)
    based on the specified stage in the pipeline. The stages can be "fit" (training), "test" (testing), or "predict" (inference).

    It performs the following steps:
        (1) Configure Paths: The YuccaConfigurator class is used to set up paths for training,
        including the training data directory, output path, and plans file path.

        (2) Create Augmentation Objects: The YuccaAugmentationComposer class is used to create data augmentation objects
        (augmenter.train_transforms and augmenter.val_transforms) based on the specified patch size and model dimensionality.

        (3) Set Up PyTorch Lightning Modules: The YuccaLightningModule class is initialized with the appropriate configuration
        using the YuccaConfigurator object. This module is responsible for defining the neural network architecture, loss functions,
        and optimization strategies.

        (4) Set Up PyTorch Lightning Data Module: The YuccaDataModule class is initialized with the composed training and validation transforms,
        along with the YuccaConfigurator object. This module handles the loading and preprocessing of the training, validation, and prediction datasets.

        (5) Initialize PyTorch Lightning Trainer: The PyTorch Lightning Trainer is configured with the necessary settings, including callbacks,
        output directory, precision, and other specified parameters.
    """

    def __init__(
        self,
        augmentation_params: dict = {},
        batch_size: Union[int, Literal["tiny"]] = None,
        ckpt_path: str = None,
        continue_from_most_recent: bool = True,
        deep_supervision: bool = False,
        enable_logging: bool = True,
        experiment: str = "default",
        learning_rate: float = 1e-3,
        loss: str = None,
        max_epochs: int = 1000,
        max_vram: int = 12,
        model_dimensions: str = "3D",
        model_name: str = "TinyUNet",
        momentum: float = 0.9,
        num_workers: Optional[int] = None,
        patch_based_training: bool = True,
        patch_size: Union[tuple, Literal["max", "min", "mean"]] = None,
        planner: str = "YuccaPlanner",
        precision: str = "bf16-mixed",
        profile: bool = False,
        split_idx: int = 0,
        split_data_method: str = "kfold",
        split_data_param: int = 5,
        step_logging: bool = False,
        task: str = None,
        train_batches_per_step: int = 250,
        use_label_regions: bool = False,
        val_batches_per_step: int = 50,
        **kwargs,
    ):
        self.ckpt_path = ckpt_path
        self.continue_from_most_recent = continue_from_most_recent
        self.deep_supervision = deep_supervision
        self.enable_logging = enable_logging
        self.experiment = experiment
        self.learning_rate = float(learning_rate)
        self.loss = loss
        self.max_epochs = max_epochs
        self.max_vram = max_vram
        self.model_dimensions = model_dimensions
        self.model_name = model_name
        self.momentum = float(momentum)
        self.name = self.__class__.__name__
        self.num_workers = num_workers
        self.augmentation_params = augmentation_params
        self.batch_size = batch_size
        self.patch_based_training = patch_based_training
        self.patch_size = patch_size
        self.planner = planner
        self.precision = precision
        self.profile = profile
        self.split_idx = split_idx
        self.split_data_method = split_data_method
        self.split_data_param = split_data_param
        self.step_logging = step_logging
        self.task = task
        self.train_batches_per_step = train_batches_per_step
        self.use_label_regions = use_label_regions
        self.val_batches_per_step = val_batches_per_step
        self.kwargs = kwargs

        # Automatically changes bfloat training if we're running on a GPU
        # that doesn't support it (otherwise it just crashes.)
        if "bf" in self.precision and torch.cuda.is_available():
            if not torch.cuda.is_bf16_supported():
                self.precision = self.precision.replace("bf", "")

        if self.kwargs.get("fast_dev_run"):
            self.setup_fast_dev_run()

        # defaults
        self.data_module_class = YuccaDataModule
        self.lightning_module = YuccaLightningModule
        self.trainer = L.Trainer
        self.train_dataset_class = YuccaTrainDataset
        self.test_dataset_class = YuccaTestDataset

    def initialize(
        self,
        stage: Literal["fit", "test", "predict"],
        disable_tta: bool = False,
        pred_include_cases: list = None,
        overwrite_predictions: bool = False,
        pred_data_dir: str = None,
        save_softmax: bool = False,
        prediction_output_dir: str = "./",
    ):
        # Here we configure the outpath we will use to store model files and metadata
        # along with the path to plans file which will also be loaded.
        task_config = get_task_config(
            task=self.task,
            continue_from_most_recent=self.continue_from_most_recent,
            manager_name=self.name,
            model_dimensions=self.model_dimensions,
            model_name=self.model_name,
            patch_based_training=self.patch_based_training,
            planner_name=self.planner,
            experiment=self.experiment,
            split_idx=self.split_idx,
            split_data_method=self.split_data_method,
            split_data_param=self.split_data_param,
        )

        path_config = get_path_config(task_config=task_config)

        self.ckpt_config = get_checkpoint_config(
            ckpt_path=self.ckpt_path,
            continue_from_most_recent=task_config.continue_from_most_recent,
            current_experiment=task_config.experiment,
            path_config=path_config,
        )

        seed_config = seed_everything_and_get_seed_config(ckpt_seed=self.ckpt_config.ckpt_seed)

        plan_config = self.get_plan_config(
            ckpt_plans=self.ckpt_config.ckpt_plans,
            plans_path=path_config.plans_path,
            stage=stage,
            use_label_regions=self.use_label_regions,
        )

        callback_config = get_callback_config(
            save_dir=path_config.save_dir,
            version_dir=path_config.version_dir,
            ckpt_version_dir=self.ckpt_config.ckpt_version_dir,
            ckpt_wandb_id=self.ckpt_config.ckpt_wandb_id,
            experiment=task_config.experiment,
            version=path_config.version,
            enable_logging=self.enable_logging,
            log_lr=True,
            prediction_output_dir=prediction_output_dir,
            profile=self.profile,
            save_softmax=save_softmax,
        )

        if stage == "fit":
            splits_config = get_split_config(task_config.split_method, task_config.split_param, path_config)
        else:
            splits_config = SplitConfig()

        input_dims_config = get_input_dims_config(
            plan=plan_config.plans,
            model_dimensions=task_config.model_dimensions,
            num_classes=plan_config.num_classes,
            ckpt_patch_size=self.ckpt_config.ckpt_patch_size,
            model_name=task_config.model_name,
            max_vram=self.max_vram,
            batch_size=self.batch_size,
            patch_based_training=task_config.patch_based_training,
            patch_size=self.patch_size,
        )

        augmenter = YuccaAugmentationComposer(
            deep_supervision=self.deep_supervision,
            patch_size=input_dims_config.patch_size,
            is_2D=True if self.model_dimensions == "2D" else False,
            parameter_dict=self.augmentation_params,
            task_type_preset=plan_config.task_type,
            label_regions=plan_config.regions_in_order if plan_config.use_label_regions else None,
        )

        self.model_module = self.lightning_module(
            config=task_config.lm_hparams()
            | path_config.lm_hparams()
            | self.ckpt_config.lm_hparams()
            | seed_config.lm_hparams()
            | splits_config.lm_hparams()
            | plan_config.lm_hparams()
            | input_dims_config.lm_hparams()
            | callback_config.lm_hparams()
            | augmenter.lm_hparams(),
            deep_supervision=self.deep_supervision,
            learning_rate=self.learning_rate,
            loss_fn=self.loss,
            momentum=self.momentum,
            step_logging=self.step_logging,
            test_time_augmentation=not disable_tta if disable_tta is True else augmenter.mirror_p_per_sample > 0,
        )

        self.data_module = self.data_module_class(
            allow_missing_modalities=plan_config.allow_missing_modalities,
            composed_train_transforms=augmenter.train_transforms,
            composed_val_transforms=augmenter.val_transforms,
            pred_include_cases=pred_include_cases,
            image_extension=plan_config.image_extension,
            input_dims_config=input_dims_config,
            overwrite_predictions=overwrite_predictions,
            num_workers=self.num_workers,
            pred_data_dir=pred_data_dir,
            pred_save_dir=prediction_output_dir,
            pre_aug_patch_size=augmenter.pre_aug_patch_size,
            splits_config=splits_config,
            split_idx=task_config.split_idx,
            task_type=plan_config.task_type,
            test_dataset_class=self.test_dataset_class,
            train_data_dir=path_config.train_data_dir,
            train_dataset_class=self.train_dataset_class,
        )

        self.verify_modules_are_valid()
        self.visualize_model_with_FLOPs(self.model_module, input_dims_config)

        self.trainer = L.Trainer(
            callbacks=callback_config.callbacks,
            default_root_dir=path_config.save_dir,
            limit_train_batches=self.train_batches_per_step,
            limit_val_batches=self.val_batches_per_step,
            log_every_n_steps=min(self.train_batches_per_step, 50),
            logger=callback_config.loggers,
            precision=self.precision,
            profiler=callback_config.profiler,
            enable_progress_bar=not self.enable_logging,
            max_epochs=self.max_epochs,
            **self.kwargs,
        )

    def run_training(self):
        self.initialize(stage="fit")
        self.trainer.fit(
            model=self.model_module,
            datamodule=self.data_module,
            ckpt_path="last",
        )
        self.finish()

    def run_finetuning(self):
        self.initialize(stage="fit")
        self.model_module.load_state_dict(
            state_dict=torch.load(self.ckpt_config.ckpt_path, map_location=torch.device("cpu"))["state_dict"], strict=False
        )
        self.trainer.fit(
            model=self.model_module,
            datamodule=self.data_module,
        )
        self.finish()

    def predict_folder(
        self,
        input_folder,
        disable_tta: bool = False,
        overwrite_predictions: bool = False,
        output_folder: str = yucca_results,
        pred_include_cases: list = None,
        save_softmax=False,
    ):
        self.batch_size = 1
        self.initialize(
            stage="predict",
            disable_tta=disable_tta,
            pred_include_cases=pred_include_cases,
            overwrite_predictions=overwrite_predictions,
            pred_data_dir=input_folder,
            prediction_output_dir=output_folder,
            save_softmax=save_softmax,
        )

        self.trainer.predict(
            model=self.model_module,
            dataloaders=self.data_module,
            ckpt_path=self.ckpt_config.ckpt_path,
            return_predictions=False,
        )
        self.finish()

    @staticmethod
    def get_plan_config(ckpt_plans, plans_path, stage, use_label_regions):
        plan_config = get_plan_config(
            ckpt_plans=ckpt_plans,
            plans_path=plans_path,
            use_label_regions=use_label_regions,
            stage=stage,
        )
        return plan_config

    def finish(self):
        wandb.finish()

    def verify_modules_are_valid(self):
        # Method to expand for additional verifications
        self.verify_samplers_are_valid()

    def verify_samplers_are_valid(self):
        if (
            not issubclass(self.data_module.train_sampler, InfiniteRandomSampler) and self.train_batches_per_step is not None
        ) or (not issubclass(self.data_module.val_sampler, InfiniteRandomSampler) and self.val_batches_per_step is not None):
            logging.info(
                "Warning: you are limiting the amount of batches pr. step, but not sampling using InfiniteRandomSampler."
            )

    def visualize_model_with_FLOPs(self, lightning_module, input_dims_config):
        try:
            flops = self.get_flops(
                lightning_module, input_dims_config.batch_size, input_dims_config.num_modalities, input_dims_config.patch_size
            )
            logging.info("\n" + flop_count_table(flops))
        except RuntimeError:
            logging.info("\n Model architecture could not be visualized.")

    @staticmethod
    def get_flops(lightning_module, batch_size, modalities, patch_size):
        data = torch.randn((batch_size, modalities, *patch_size))
        flops = measure_FLOPs(lightning_module.model, data)
        return flops

    def setup_fast_dev_run(self):
        self.batch_size = 2
        self.patch_size = (32, 32)
        self.enable_logging = False
        self.model_dimensions = "2D"
        self.precision = 32
        self.kwargs["accelerator"] = "cpu"
        self.train_batches_per_step = 10
        self.val_batches_per_step = 5


if __name__ == "__main__":
    # path = "/home/zcr545/YuccaData/yucca_models/Task001_OASIS/UNet__3D/YuccaPlanner/YuccaLightningManager/0/2023_11_08_15_19_14/checkpoints/test_ckpt.ckpt"
    path = None
    Manager = YuccaManager(
        disable_logging=False,
        ckpt_path=path,
        model_name="TinyUNet",
        model_dimensions="2D",
        num_workers=0,
        split_idx=0,
        step_logging=True,
        task="Task001_OASIS",
    )
    Manager.initialize("fit")
    # Manager.run_training()
    # Manager.predict_folder(
    #    input_folder="/home/zcr545/YuccaData/yucca_raw_data/Task001_OASIS/imagesTs",
    #    output_folder="/home/zcr545/YuccaData/yucca_predictions",
    # )



import torch.nn as nn


class DepthwiseSeparableConv(nn.Module):
    def __init__(
        self,
        input_channels,
        output_channels,
        conv_op=nn.Conv2d,
        kernel_size=3,
        stride=1,
        padding=1,
        dilation=1,
        bias=True,
    ):
        super().__init__()
        self.depthconv = conv_op(
            input_channels,
            input_channels,
            kernel_size,
            groups=input_channels,
            stride=stride,
            padding=padding,
            dilation=dilation,
            bias=bias,
        )
        self.pointwiseconv = conv_op(input_channels, output_channels, kernel_size=1)


class ConvDropoutNormNonlin(nn.Module):
    """
    2D Convolutional layers
    Arguments:
    num_in_filters {int} -- number of input filters
    num_out_filters {int} -- number of output filters
    kernel_size {tuple} -- size of the convolving kernel
    stride {tuple} -- stride of the convolution (default: {(1, 1)})
    activation {str} -- activation function (default: {'relu'})
    """

    def __init__(
        self,
        input_channels,
        output_channels,
        conv_op=nn.Conv2d,
        conv_kwargs={
            "kernel_size": 3,
            "stride": 1,
            "padding": 1,
            "dilation": 1,
            "bias": True,
        },
        norm_op=nn.BatchNorm2d,
        norm_op_kwargs={"eps": 1e-5, "affine": True, "momentum": 0.1},
        dropout_op=nn.Dropout2d,
        dropout_op_kwargs={"p": 0.5, "inplace": True},
        nonlin=nn.LeakyReLU,
        nonlin_kwargs={"negative_slope": 1e-2, "inplace": True},
    ):
        super().__init__()

        self.nonlin_kwargs = nonlin_kwargs
        self.nonlin = nonlin
        self.dropout_op = dropout_op
        self.dropout_op_kwargs = dropout_op_kwargs
        self.norm_op_kwargs = norm_op_kwargs
        self.conv_kwargs = conv_kwargs
        self.conv_op = conv_op
        self.norm_op = norm_op

        self.conv = self.conv_op(input_channels, output_channels, **self.conv_kwargs)

        if self.dropout_op is not None and self.dropout_op_kwargs["p"] is not None and self.dropout_op_kwargs["p"] > 0:
            self.dropout = self.dropout_op(**self.dropout_op_kwargs)
        else:
            self.dropout = None
        self.norm = self.norm_op(output_channels, **self.norm_op_kwargs)
        self.activation = self.nonlin(**self.nonlin_kwargs)

    def forward(self, x):
        x = self.conv(x)
        if self.dropout is not None:
            x = self.dropout(x)
        return self.activation(self.norm(x))


class ConvDropoutNorm(ConvDropoutNormNonlin):
    def forward(self, x):
        x = self.conv(x)
        if self.dropout is not None:
            x = self.dropout(x)
        return self.norm(x)


class DoubleConvDropoutNormNonlin(nn.Module):
    """
    2D Convolutional layers
    Arguments:
    num_in_filters {int} -- number of input filters
    num_out_filters {int} -- number of output filters
    kernel_size {tuple} -- size of the convolving kernel
    stride {tuple} -- stride of the convolution (default: {(1, 1)})
    activation {str} -- activation function (default: {'relu'})
    """

    def __init__(
        self,
        input_channels,
        output_channels,
        conv_op=nn.Conv2d,
        conv_kwargs={
            "kernel_size": 3,
            "stride": 1,
            "padding": 1,
            "dilation": 1,
            "bias": True,
        },
        norm_op=nn.BatchNorm2d,
        norm_op_kwargs={"eps": 1e-5, "affine": True, "momentum": 0.1},
        dropout_op=nn.Dropout2d,
        dropout_op_kwargs={"p": 0.5, "inplace": True},
        nonlin=nn.LeakyReLU,
        nonlin_kwargs={"negative_slope": 1e-2, "inplace": True},
    ):
        super().__init__()

        self.nonlin_kwargs = nonlin_kwargs
        self.nonlin = nonlin
        self.dropout_op = dropout_op
        self.dropout_op_kwargs = dropout_op_kwargs
        self.norm_op_kwargs = norm_op_kwargs
        self.conv_kwargs = conv_kwargs
        self.conv_op = conv_op
        self.norm_op = norm_op

        self.conv1 = ConvDropoutNormNonlin(
            input_channels,
            output_channels,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
        )
        self.conv2 = ConvDropoutNormNonlin(
            output_channels,
            output_channels,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
        )

    def forward(self, x):
        x = self.conv1(x)
        x = self.conv2(x)
        return x



from yucca.data.augmentation.transforms.YuccaTransform import YuccaTransform
from yucca.functional.transforms import augment_gamma
import numpy as np


class Gamma(YuccaTransform):
    """
    WRAPPER FOR NNUNET AUGMENT GAMMA: https://github.com/MIC-DKFZ/batchgenerators/blob/8822a08a7dbfa4986db014e6a74b040778164ca6/batchgenerators/augmentations/color_augmentations.py

    Augments by changing 'gamma' of the image (same as gamma correction in photos or computer monitors

    :param gamma_range: range to sample gamma from. If one value is smaller than 1 and the other one is
    larger then half the samples will have gamma <1 and the other >1 (in the inverval that was specified).
    Tuple of float. If one value is < 1 and the other > 1 then half the images will be augmented with gamma values
    smaller than 1 and the other half with > 1
    :param invert_image: whether to invert the image before applying gamma augmentation
    :param retain_stats: Gamma transformation will alter the mean and std of the data in the patch. If retain_stats=True,
    the data will be transformed to match the mean and standard deviation before gamma augmentation. retain_stats
    can also be callable (signature retain_stats() -> bool)
    """

    def __init__(
        self,
        data_key="image",
        p_per_sample: float = 1.0,
        p_invert_image: float = 0.05,
        gamma_range=(0.5, 2.0),
        per_channel=True,
        clip_to_input_range=False,
    ):
        self.data_key = data_key
        self.p_per_sample = p_per_sample
        self.gamma_range = gamma_range
        self.p_invert_image = p_invert_image
        self.per_channel = per_channel
        self.clip_to_input_range = clip_to_input_range

    @staticmethod
    def get_params(p_invert_image):
        # No parameters to retrieve
        do_invert = False
        if np.random.uniform() < p_invert_image:
            do_invert = True
        return do_invert

    def __gamma__(self, image, gamma_range, invert_image, per_channel):
        return augment_gamma(
            image,
            gamma_range,
            invert_image,
            per_channel,
            clip_to_input_range=self.clip_to_input_range,
        )

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape.\
            \nShould be (b, c, x, y, z) or (b, c, x, y) and is: {data_dict[self.data_key].shape}"

        for b in range(data_dict[self.data_key].shape[0]):
            if np.random.uniform() < self.p_per_sample:
                do_invert = self.get_params(self.p_invert_image)
                data_dict[self.data_key][b] = self.__gamma__(
                    data_dict[self.data_key][b],
                    self.gamma_range,
                    do_invert,
                    per_channel=self.per_channel,
                )
        return data_dict



import shutil
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles, subdirs
from yucca.pipeline.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data, yucca_source


def convert(path: str = yucca_source, subdir: str = "ACDC"):
    """INPUT DATA - Define input path and suffixes"""
    path = join(path, subdir)
    file_suffix = ".nii.gz"

    """ OUTPUT DATA - Define the task name and prefix """
    task_name = "Task016_ACDC"
    task_prefix = "ACDC"

    """ Access the input data. If images are not split into train/test, and you wish to randomly 
    split the data, uncomment and adapt the following lines to fit your local path. """

    images_dir_tr = labels_dir_tr = join(path, "database", "training")
    images_dir_ts = labels_dir_ts = join(path, "database", "testing")

    train_samples = subdirs(images_dir_tr, join=False)
    test_samples = subdirs(images_dir_ts, join=False)

    """ Then define target paths """
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    """Populate Target Directory
    This is also the place to apply any re-orientation, resampling and/or label correction."""

    for patient in train_samples:
        for sTr in subfiles(join(images_dir_tr, patient), join=False, suffix=file_suffix):
            if "_4d" in sTr or "_gt" in sTr:
                # skip these. We'll infer the label path from the remaining subjects
                continue
            case_id = sTr[: -len(file_suffix)]
            src_image_file_path = join(images_dir_tr, patient, case_id + file_suffix)
            src_label_file_path = join(labels_dir_tr, patient, case_id + "_gt" + file_suffix)
            dst_image_file_path = f"{target_imagesTr}/{task_prefix}_{case_id}_000.nii.gz"
            dst_label_file_path = f"{target_labelsTr}/{task_prefix}_{case_id}.nii.gz"

            shutil.copy2(src_image_file_path, dst_image_file_path)
            shutil.copy2(src_label_file_path, dst_label_file_path)
    del sTr

    for patient in test_samples:
        for sTs in subfiles(join(images_dir_ts, patient), join=False, suffix=file_suffix):
            if "_4d" in sTs or "_gt" in sTs:
                # skip these. We'll infer the label path from the remaining subjects
                continue
            case_id = sTs[: -len(file_suffix)]
            src_image_file_path = join(images_dir_ts, patient, case_id + file_suffix)
            src_label_file_path = join(labels_dir_ts, patient, case_id + "_gt" + file_suffix)
            dst_image_file_path = f"{target_imagesTs}/{task_prefix}_{case_id}_000.nii.gz"
            dst_label_file_path = f"{target_labelsTs}/{task_prefix}_{case_id}.nii.gz"

            shutil.copy2(src_image_file_path, dst_image_file_path)
            shutil.copy2(src_label_file_path, dst_label_file_path)

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        modalities=("cine-MRI",),
        labels={
            0: "background",
            1: "Right Ventricle Cavity",
            2: "Myocardium",
            3: "Left Ventricle Cavity",
        },
        dataset_name=task_name,
        license="",
        dataset_description="ACDC - Automated Cardiac Diagnosis Challenge",
        dataset_reference="https://www.creatis.insa-lyon.fr/Challenge/acdc/index.html",
    )


if __name__ == "__main__":
    convert()



import nibabel as nib
import shutil
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.pipeline.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data
from yucca.functional.testing.data.nifti import verify_spacing_is_equal, verify_orientation_is_equal

# INPUT DATA
# Define input path and extension

folder_with_images = "/home/zcr545/data/data/public_datasets/decathlon/Task10_Colon"
file_extension = ".nii.gz"

# OUTPUT DATA
# Define the task name and prefix
task_name = "Task040_MSD_Colon"
task_prefix = ""

# Set target paths
target_base = join(yucca_raw_data, task_name)
target_imagesTr = join(target_base, "imagesTr")
target_labelsTr = join(target_base, "labelsTr")
target_imagesTs = join(target_base, "imagesTs")
target_labelsTs = join(target_base, "labelsTs")

maybe_mkdir_p(target_imagesTr)
maybe_mkdir_p(target_labelsTs)
maybe_mkdir_p(target_imagesTs)
maybe_mkdir_p(target_labelsTr)

# Split data
images_dir_tr = join(folder_with_images, "imagesTr")
labels_dir_tr = join(folder_with_images, "labelsTr")
images_dir_ts = join(folder_with_images, "imagesTs")

# Populate Target Directory
# This is also the place to apply any re-orientation, resampling and/or label correction.

for sTr in subfiles(images_dir_tr, join=False):
    image_path = join(images_dir_tr, sTr)
    label_path = join(labels_dir_tr, sTr)
    sTr = sTr[: -len(file_extension)]

    image = nib.load(image_path)
    label = nib.load(label_path)
    assert verify_spacing_is_equal(image, label), "spacing"
    assert verify_orientation_is_equal(image, label), "orientation"

    shutil.copy2(image_path, f"{target_imagesTr}/{sTr}_000.nii.gz")
    shutil.copy2(label_path, f"{target_labelsTr}/{sTr}.nii.gz")

for sTs in subfiles(images_dir_ts, join=False):
    image_path = join(images_dir_ts, sTs)
    sTs = sTs[: -len(file_extension)]

    image = nib.load(image_path)

    shutil.copy2(image_path, f"{target_imagesTs}/{sTs}_000.nii.gz")

generate_dataset_json(
    join(target_base, "dataset.json"),
    target_imagesTr,
    target_imagesTs,
    modalities=("CT",),
    labels={0: "Background", 1: "Colon Cancer Primaries"},
    dataset_name=task_name,
    license="CC-BY-SA 4.0",
    dataset_description="Decathlon: Colon",
    dataset_reference="King's College London",
)



from torch import Tensor
from torchmetrics.classification import MulticlassF1Score
from torchmetrics.classification import Accuracy as TorchAccuracy
from torchmetrics.classification import AUROC as TorchAUROC
from torchmetrics import Metric


class F1(MulticlassF1Score):
    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        if len(target.shape) == len(input.shape):
            assert target.shape[1] == 1
            target = target[:, 0]
        return super().forward(input, target)


class AUROC(Metric):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.method = TorchAUROC(*args, **kwargs)

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        if len(target.shape) == len(input.shape):
            assert target.shape[1] == 1
            target = target[:, 0]
        return self.method(input, target.long())

    def compute(self):
        pass

    def update(self):
        pass


class Accuracy(Metric):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.method = TorchAccuracy(*args, **kwargs)

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        if len(target.shape) == len(input.shape):
            assert target.shape[1] == 1
            target = target[:, 0]
        return self.method(input, target.long())

    def compute(self):
        pass

    def update(self):
        pass



from yucca.pipeline.managers.YuccaManagerV2 import YuccaManagerV2
from yucca.lightning_modules.YuccaLightningModule_onehot_labels import YuccaLightningModule_onehot_labels


class YuccaManagerV2_labelregions(YuccaManagerV2):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(
            *args,
            **kwargs,
        )
        self.lightning_module = YuccaLightningModule_onehot_labels
        self.use_label_regions = True



import re
from yucca.pipeline.preprocessing.YuccaPreprocessor import YuccaPreprocessor
from yucca.paths import yucca_preprocessed_data, yucca_raw_data
from batchgenerators.utilities.file_and_folder_operations import (
    join,
    subfiles,
)


class UnsupervisedPreprocessor(YuccaPreprocessor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # set up for self-/unsupervised
        self.classification = False
        self.label_exists = False
        self.preprocess_label = False

    def initialize_paths(self):
        # Have to overwrite how we get the subject_ids as there's no labelsTr to get them from.
        # Therefore we use the imagesTr folder and remove the modality suffix.
        self.target_dir = join(yucca_preprocessed_data, self.task, self.plans["plans_name"])
        self.input_dir = join(yucca_raw_data, self.task)
        self.imagepaths = subfiles(join(self.input_dir, "imagesTr"), suffix=self.image_extension)

        subject_ids = subfiles(join(self.input_dir, "imagesTr"), suffix=self.image_extension, join=False)
        self.subject_ids = [re.sub(r"_\d+\.", ".", subject) for subject in subject_ids]



from yucca.data.augmentation.transforms.BiasField import BiasField
from yucca.data.augmentation.transforms.Blur import Blur
from yucca.data.augmentation.transforms.convert_labels_to_regions import ConvertLabelsToRegions
from yucca.data.augmentation.transforms.copy_image_to_label import CopyImageToLabel
from yucca.data.augmentation.transforms.cropping_and_padding import CropPad
from yucca.data.augmentation.transforms.formatting import (
    RemoveBatchDimension,
    RemoveSegChannelAxis,
    NumpyToTorch,
    AddBatchDimension,
    CollectMetadata,
)
from yucca.data.augmentation.transforms.Gamma import Gamma
from yucca.data.augmentation.transforms.Ghosting import MotionGhosting
from yucca.data.augmentation.transforms.Masking import Masking
from yucca.data.augmentation.transforms.Mirror import Mirror
from yucca.data.augmentation.transforms.Noise import AdditiveNoise, MultiplicativeNoise
from yucca.data.augmentation.transforms.normalize import Normalize
from yucca.data.augmentation.transforms.Ringing import GibbsRinging
from yucca.data.augmentation.transforms.sampling import DownsampleSegForDS
from yucca.data.augmentation.transforms.SimulateLowres import SimulateLowres
from yucca.data.augmentation.transforms.Spatial import Spatial



import numpy as np
import torch


def convert_labels_to_regions(data, regions):
    b, c, *shape = data.shape
    assert c == 1, "# Channels is not 1. Make sure the input to this function is a segmentation map of dims (b,c,h,w[,d])"
    region_canvas = np.zeros((b, len(regions), *shape))
    for channel, region in enumerate(regions):
        region_canvas[:, channel] = np.isin(data[:, 0], region)
    region_canvas = region_canvas.astype(np.uint8)
    return region_canvas


def convert_regions_to_labels(data, region_labels):
    b, _, *shape = data.shape
    region_canvas = np.zeros((b, 1, *shape))
    for channel, region_label in enumerate(region_labels):
        mask = data[:, channel] > 0.5
        region_canvas[:, 0][mask] = region_label
    region_canvas = region_canvas
    return region_canvas


def torch_convert_regions_to_labels(data, region_labels):
    b, _, *shape = data.shape
    region_canvas = torch.zeros((b, 1, *shape))
    for channel, region_label in enumerate(region_labels):
        mask = data[:, channel] > 0.5
        region_canvas[:, 0][mask] = region_label
    region_canvas = region_canvas
    return region_canvas



from yucca.pipeline.managers.YuccaManager import YuccaManager
from torch import optim


class YuccaManager_AdamW(YuccaManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.optim = optim.AdamW



"""
[INTERNAL]

Task_conversion file for the SONAI data, without labels and with 2D jpg images (cache).
"""

import shutil
from tqdm import tqdm
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p
from yucca.paths import yucca_raw_data
from yucca.pipeline.task_conversion.utils import generate_dataset_json


def convert(path: str, txt_file_prefix: str = "data"):
    """
    Converts the SONAI image chache to the Yucca format.

    We are abusing the subdir argument to specify a file with the label paths.

    Usage:
        python run_task_conversion.py -t Task901_SONAI_NoLabel -p /path/to/SONAI -d data

        there need to be data_train.txt and data_test.txt files in the specified directory. (data is the default prefix)
    """

    # INPUT DATA
    images_tr_txt = join(path, f"{txt_file_prefix}_train.txt")
    images_ts_txt = join(path, f"{txt_file_prefix}_test.txt")

    # Image paths are in a txt that we read line by line
    with open(images_tr_txt, "r") as f:
        image_paths_tr = f.read().splitlines()

    with open(images_ts_txt, "r") as f:
        image_paths_ts = f.read().splitlines()

    # OUTPUT DATA
    # Target names
    task_name = "Task901_SONAI_NoLabel"
    prefix = "SONAI"

    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_imagesTs = join(target_base, "imagesTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_imagesTs)

    # Populate Target Directory

    for image_path in tqdm(image_paths_tr, desc="Train"):
        target_file_name = image_path.split("/")[-1][0:-4]  # filename without extension
        target_file_name = target_file_name.replace(".", "_")  # replace dots with underscores
        shutil.copyfile(image_path, f"{target_imagesTr}/{prefix}_{target_file_name}_000.jpg")

    for image_path in tqdm(image_paths_ts, desc="Test"):
        target_file_name = image_path.split("/")[-1][0:-4]
        target_file_name = target_file_name.replace(".", "_")
        shutil.copyfile(image_path, f"{target_imagesTs}/{prefix}_{target_file_name}_000.jpg")

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        ("US2D",),
        labels={},
        dataset_name=task_name,
        license="hands off!",
        dataset_description="SONAI Dataset from image cache",
        dataset_reference="",
    )



from yucca.pipeline.managers.YuccaManager import YuccaManager


class YuccaManager_1e5(YuccaManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.learning_rate = 1e-5



from copy import deepcopy
import numpy as np


def get_bbox_for_foreground(array, background_label=0):
    array = deepcopy(array)
    array[array > background_label] = 1
    return get_bbox_for_label(array, label=1)


def get_bbox_for_label(array, label, padding=0):
    a = np.where(array == label)

    x1, x2, y1, y2 = np.min(a[0]), np.max(a[0]), np.min(a[1]), np.max(a[1])

    xmin = min(x1, x2) - padding
    xmax = max(x1, x2) + padding

    ymin = min(y1, y2) - padding
    ymax = max(y1, y2) + padding

    if len(array.shape) == 3:
        z1, z2 = np.min(a[2]), np.max(a[2])

        zmin = min(z1, z2) - padding
        zmax = max(z1, z2) + padding

        return [xmin, xmax + 1, ymin, ymax + 1, zmin, zmax + 1]

    return [xmin, xmax + 1, ymin, ymax + 1]



from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subdirs
from yucca.pipeline.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data
from sklearn.model_selection import train_test_split
import nibabel as nib
import numpy as np
import shutil
from tqdm import tqdm


def convert(path: str, subdir: str = "brats21/training_data"):
    # Target names
    task_name = "Task042_BraTS21_t1ce"
    task_prefix = "BraTS21"

    ###OUTPUT DATA
    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    # INPUT DATA
    # Input path and names
    base_in = join(path, subdir)
    file_suffix = ".nii.gz"

    # Train/Test Splits
    # We only use the train folder, as the test folder does not contain segmentations
    # to obtain those, submission to the challenge is required (it's from 2008, forget it)
    training_samples, test_samples = train_test_split(subdirs(base_in, join=False), random_state=4215532)

    ###Populate Target Directory###
    for sTr in tqdm(training_samples, desc="training"):
        src_image_file_path1 = join(base_in, sTr, sTr + "_t1ce" + file_suffix)
        dst_image_file_path1 = f"{target_imagesTr}/{task_prefix}_{sTr}_000.nii.gz"

        dst_label_path = f"{target_labelsTr}/{task_prefix}_{sTr}.nii.gz"
        label = nib.load(join(base_in, sTr, sTr + "_seg" + file_suffix))
        labelarr = label.get_fdata()
        labelarr[labelarr == 4.0] = 3.0
        assert np.all(np.isin(np.unique(labelarr), np.array([0, 1, 2, 3])))
        labelnew = nib.Nifti1Image(labelarr, label.affine, label.header, dtype=np.float32)
        nib.save(labelnew, dst_label_path)

        shutil.copy2(src_image_file_path1, dst_image_file_path1)

    for sTs in tqdm(test_samples, desc="test"):
        src_image_file_path1 = join(base_in, sTs, sTs + "_t1ce" + file_suffix)
        dst_image_file_path1 = f"{target_imagesTs}/{task_prefix}_{sTs}_000.nii.gz"

        dst_label_path = f"{target_labelsTs}/{task_prefix}_{sTs}.nii.gz"
        label = nib.load(join(base_in, sTs, sTs + "_seg" + file_suffix))
        labelarr = label.get_fdata()
        labelarr[labelarr == 4.0] = 3.0
        assert np.all(np.isin(np.unique(labelarr), np.array([0, 1, 2, 3])))
        labelnew = nib.Nifti1Image(labelarr, label.affine, label.header, dtype=np.float32)
        nib.save(labelnew, dst_label_path)

        shutil.copy2(src_image_file_path1, dst_image_file_path1)

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        ("T1ce"),
        labels={0: "background", 1: "necrotic tumor core", 2: "peritumoral edematous/invaded tissue", 3: "GD-enhancing tumor"},
        dataset_name=task_name,
        license="hands off!",
        dataset_description="BraTS21",
        dataset_reference="https://www.nitrc.org/projects/msseg, https://arxiv.org/abs/2206.06694",
    )


if __name__ == "__main__":
    convert()



"""
PLEASE READ YUCCA/DOCUMENTATION/TUTORIALS/ENVIRONMENT_VARIABLES.MD FOR INFORMATION ON HOW TO SET THIS UP
"""

import os
import warnings
from dotenv import load_dotenv

from batchgenerators.utilities.file_and_folder_operations import maybe_mkdir_p

load_dotenv()

vars = ["YUCCA_SOURCE", "YUCCA_RAW_DATA", "YUCCA_PREPROCESSED_DATA", "YUCCA_MODELS", "YUCCA_RESULTS"]
vals = {}

for var in vars:
    if var in os.environ.keys():
        vals[var] = os.environ[var]
        maybe_mkdir_p(vals[var])
    else:
        warnings.warn(f"Missing environment variable {var}.")
        vals[var] = None

yucca_source = vals["YUCCA_SOURCE"]
yucca_raw_data = vals["YUCCA_RAW_DATA"]
yucca_preprocessed_data = vals["YUCCA_PREPROCESSED_DATA"]
yucca_models = vals["YUCCA_MODELS"]
yucca_results = vals["YUCCA_RESULTS"]

if "YUCCA_WANDB_ENTITY" in os.environ.keys():
    yucca_wandb_entity = os.environ["YUCCA_WANDB_ENTITY"]
else:
    yucca_wandb_entity = None



import torch
import torch.nn as nn
from yucca.networks.networks.YuccaNet import YuccaNet
from yucca.networks.blocks_and_layers.conv_blocks import Bottleneck
from torch import Tensor


class ResNet50(YuccaNet):
    """ """

    def __init__(
        self,
        input_channels: int,
        starting_filters: int = 64,
        block=Bottleneck,
        layers=[3, 4, 6, 3],
        num_classes: int = 1000,
        zero_init_residual: bool = False,
        groups: int = 1,
        width_per_group: int = 64,
        replace_stride_with_dilation=[False, False, True],
        norm_layer=nn.BatchNorm2d,
    ) -> None:
        super().__init__()
        self._norm_layer = norm_layer
        self.groups = groups
        self.num_classes = num_classes
        self.base_width = width_per_group

        self.inplanes = starting_filters
        self.dilation = 1

        self.conv1 = nn.Conv2d(
            input_channels,
            self.inplanes,
            kernel_size=7,
            stride=2,
            padding=3,
            bias=False,
        )
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2, dilate=replace_stride_with_dilation[0])
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2, dilate=replace_stride_with_dilation[1])
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2, dilate=replace_stride_with_dilation[2])
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

        self.fc = nn.Linear(512 * block.expansion, self.num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck) and m.bn3.weight is not None:
                    nn.init.constant_(m.bn3.weight, 0)  # type: ignore[arg-type]

    def _make_layer(
        self,
        block: Bottleneck,
        planes: int,
        blocks: int,
        stride: int = 1,
        dilate: bool = False,
    ) -> nn.Sequential:
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(
                    self.inplanes,
                    planes * block.expansion,
                    kernel_size=1,
                    stride=stride,
                    bias=False,
                ),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride,
                downsample,
                self.groups,
                self.base_width,
                previous_dilation,
                norm_layer,
            )
        )
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=norm_layer,
                )
            )

        return nn.Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        # See note [TorchScript super()]
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        low_level_features = self.layer1(x)  # Potential Skip Connection 1 Here
        # Outputs of layer1 are also known as "low_level" features for DeepLabV3+

        x = self.layer2(low_level_features)  # Potential Skip Connection 2 Here
        x = self.layer3(x)  # Potential Skip Connection 3 Here
        high_level_features = self.layer4(x)  # Potential Skip Connection 4 Here
        # Outputs of layer1 are also known as "out" features for DeepLabV3+

        x = self.avgpool(high_level_features)
        x = torch.flatten(x, 1)
        x = self.fc(x)

        return x



import numpy as np


def mask_batch(batch, pixel_value, ratio, token_size):
    assert len(batch.shape[2:]) == len(token_size), (
        "mask token size not compatible with input data" f"mask token is: {token_size} and image is shape: {batch.shape[2:]}"
    )
    # np.ceil to get a grid with exact or larger dims than the input image
    # later we will crop it to the desired dims
    slices = [slice(0, shape) for shape in batch.shape[2:]]
    grid_dims = np.ceil(batch.shape[2:] / np.array(token_size)).astype(np.uint8)

    grid_flat = np.ones(np.prod(grid_dims))
    grid_flat[: int(len(grid_flat) * ratio)] = 0
    np.random.shuffle(grid_flat)
    grid = grid_flat.reshape(grid_dims)
    for idx, size in enumerate(token_size):
        grid = np.repeat(grid, repeats=size, axis=idx)

    batch[:, :, grid[*slices] == 0] = pixel_value
    return batch



import numpy as np
from scipy.ndimage import gaussian_filter


def blur(image, sigma, clip_to_input_range):
    img_min = image.min()
    img_max = image.max()

    image = gaussian_filter(image, sigma, order=0)
    if clip_to_input_range:
        image = np.clip(image, a_min=img_min, a_max=img_max)

    return image



import numpy as np
import logging
import matplotlib.pyplot as plt


def get_train_fig_with_inp_out_tar(input, output, target, fig_title, task_type: str = "segmentation"):
    # This needs to handle the following cases:
    # Segmentation      : {"input": (m,x,y(,z)), "target": (1,x,y(,z)), "output": (c,x,y(,z))}
    # Self-supervised   : {"input": (m,x,y(,z)), "target": (m,x,y(,z)), "output": (m,x,y(,z))}
    # Classification    : {"input": (m,x,y(,z)), "target": (1,x), "output": (c,x)}

    channel_idx = np.random.randint(0, input.shape[0])

    if len(input.shape) == 4:  # 3D images.
        # We need to select a slice to visualize.
        if task_type == "segmentation" and len(target[0].nonzero()[0]) > 0:
            # Select a foreground slice if any exist.
            foreground_locations = target[0].nonzero()
            slice_to_visualize = foreground_locations[0][np.random.randint(0, len(foreground_locations[0]))]
        else:
            slice_to_visualize = np.random.randint(0, input.shape[1])

        input = input[:, slice_to_visualize]
        if len(target.shape) == 4:
            target = target[:, slice_to_visualize]
        if len(output.shape) == 4:
            output = output[:, slice_to_visualize]

    image = input[channel_idx]

    if task_type in ["segmentation", "classification"]:
        target = target[0]
        output = output.argmax(0)
    elif task_type == "self-supervised":
        target = target[channel_idx]
        output = output[channel_idx]
    else:
        logging.warn(
            f"Unknown task type. Found {task_type} and expected one in ['classification', 'segmentation', 'self-supervised']"
        )

    if len(target.shape) == 1:
        fig, axes = plt.subplots(nrows=1, ncols=1, figsize=(5, 5), dpi=100, constrained_layout=True)
        axes[0].imshow(image, cmap="gray", vmin=np.quantile(image, 0.01), vmax=np.quantile(image, 0.99))
        axes[0].set_title("input")
        fig.suptitle(f"{fig_title}. Target: {target} | Output: {output}", fontsize=16)
    else:
        fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(15, 5), dpi=100, constrained_layout=True)
        axes[0].imshow(image, cmap="gray", vmin=np.quantile(image, 0.01), vmax=np.quantile(image, 0.99))
        axes[0].set_title("input")
        axes[1].imshow(target, cmap="gray")
        axes[1].set_title("target")
        axes[2].imshow(output, cmap="gray")
        axes[2].set_title("output")
        fig.suptitle(fig_title, fontsize=16)
    return fig



from yucca.pipeline.managers.YuccaManager import YuccaManager
from yucca.data.augmentation.augmentation_presets import CT


class YuccaManager_CT(YuccaManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.augmentation_params = CT



"""
Negative log likelihood loss wrapper
Makes loss function robust to target tensors with a channel dimension
"""

from torch import nn, Tensor


class CE(nn.CrossEntropyLoss):
    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        if len(target.shape) == len(input.shape):
            assert target.shape[1] == 1
            target = target[:, 0]
        return super().forward(input, target.long())


class BCEWithLogitsLoss(nn.BCEWithLogitsLoss):
    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        return super().forward(input, target.float())



import shutil
import gzip
from sklearn.model_selection import train_test_split
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.pipeline.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data


def convert(path: str, subdir: str = "MyDataset"):
    """INPUT DATA - Define input path and suffixes"""
    path = join(path, subdir)
    file_suffix = ".nii"

    """ OUTPUT DATA - Define the task name and prefix """
    task_name = "Task000_MyTask"
    task_prefix = "MyTask"

    """ Access the input data. If images are not split into train/test, and you wish to randomly 
    split the data, uncomment and adapt the following lines to fit your local path. """

    images_dir = join(path, "data_dir", "images")
    labels_dir = join(path, "data_dir", "labels")

    samples = subfiles(labels_dir, join=False, suffix=file_suffix)
    train_samples, test_samples = train_test_split(samples, test_size=0.2, random_state=42154)

    images_dir_tr = images_dir_ts = images_dir
    labels_dir_tr = labels_dir_ts = labels_dir

    """ If images are already split into train/test and images/labels uncomment and adapt the following 
    lines to fit your local path."""

    # images_dir_tr = join(path, 'train_dir', 'images')
    # labels_dir_tr = join(path, 'train_dir', 'labels')
    # train_samples = subfiles(labels_dir_tr, join=False, suffix=file_suffix)

    # images_dir_ts = join(path, 'test_dir', 'images')
    # labels_dir_ts = join(path, 'test_dir', 'labels')
    # test_samples = subfiles(labels_dir_ts, join=False, suffix=file_suffix)

    """ Then define target paths """
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    """Populate Target Directory
    This is also the place to apply any re-orientation, resampling and/or label correction."""

    for sTr in train_samples:
        case_id = sTr[: -len(file_suffix)]
        image_file = open(join(images_dir_tr, sTr), "rb")
        label = open(join(labels_dir_tr, sTr), "rb")
        shutil.copyfileobj(image_file, gzip.open(f"{target_imagesTr}/{task_prefix}_{case_id}_000.nii.gz", "wb"))
        shutil.copyfileobj(label, gzip.open(f"{target_labelsTr}/{task_prefix}_{case_id}.nii.gz", "wb"))

    for sTs in test_samples:
        case_id = sTs[: -len(file_suffix)]
        image_file = open(join(images_dir_ts, sTs), "rb")
        label = open(join(labels_dir_ts, sTs), "rb")
        shutil.copyfileobj(image_file, gzip.open(f"{target_imagesTs}/{task_prefix}_{case_id}_000.nii.gz", "wb"))
        shutil.copyfileobj(label, gzip.open(f"{target_labelsTs}/{task_prefix}_{case_id}.nii.gz", "wb"))

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        modalities=("T1",),
        labels={0: "background, probably", 1: "Fake Label", 2: "Also Fake Label"},
        dataset_name=task_name,
        license="Template",
        dataset_description="Template Dataset",
        dataset_reference="Link to source or similar",
    )



{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "import lightning as L\n",
    "import os\n",
    "from batchgenerators.utilities.file_and_folder_operations import load_json\n",
    "from yucca.pipeline.managers.YuccaManager import YuccaManager\n",
    "from yucca.paths import yucca_raw_data, yucca_preprocessed_data, yucca_models\n",
    "from yucca.pipeline.configuration.configure_task import TaskConfig\n",
    "from yucca.pipeline.configuration.configure_paths import get_path_config\n",
    "from yucca.pipeline.configuration.configure_callbacks import get_callback_config\n",
    "from yucca.pipeline.configuration.split_data import get_split_config\n",
    "from yucca.pipeline.configuration.configure_input_dims import InputDimensionsConfig\n",
    "from yucca.data.augmentation.YuccaAugmentationComposer import YuccaAugmentationComposer\n",
    "from yucca.data.augmentation.augmentation_presets import generic\n",
    "from yucca.lightning_modules.YuccaLightningModule import YuccaLightningModule\n",
    "from yucca.data.data_modules.YuccaDataModule import YuccaDataModule"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Set some variables that we'll need"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [],
   "source": [
    "config = {\n",
    "    \"batch_size\": 2,\n",
    "    \"dims\": \"2D\",\n",
    "    \"deep_supervision\": False,\n",
    "    \"experiment\": \"default\",\n",
    "    \"learning_rate\": 1e-3,\n",
    "    \"loss_fn\": \"DiceCE\",\n",
    "    \"model_name\": \"TinyUNet\",\n",
    "    \"momentum\": 0.99,\n",
    "    \"num_classes\": 3,\n",
    "    \"num_modalities\": 1,\n",
    "    \"patch_size\": (32, 32),\n",
    "    \"plans_name\": \"demo\",\n",
    "    \"plans\": None,\n",
    "    \"split_idx\": 0,\n",
    "    \"split_method\": \"kfold\",\n",
    "    \"split_param\": 5,\n",
    "    \"task\": \"Task001_OASIS\",\n",
    "    \"task_type\": \"segmentation\",\n",
    "}"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "WARNING:root:Reusing already computed split file which was split using the kfold method and parameter 5.\n",
      "INFO:root:YuccaLightningModule initialized with the following config: {'batch_size': 2, 'dims': '2D', 'deep_supervision': False, 'experiment': 'default', 'learning_rate': 0.001, 'loss_fn': 'DiceCE', 'model_name': 'TinyUNet', 'momentum': 0.99, 'num_classes': 3, 'num_modalities': 1, 'patch_size': (32, 32), 'plans_name': 'demo', 'plans': None, 'split_idx': 0, 'split_method': 'kfold', 'split_param': 5, 'task': 'Task001_OASIS', 'task_type': 'segmentation', 'continue_from_most_recent': True, 'manager_name': '', 'model_dimensions': '2D', 'patch_based_training': True, 'planner_name': 'demo', 'plans_path': '/Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo/demo_plans.json', 'save_dir': '/Users/zcr545/Desktop/Projects/repos/yucca_data/models/Task001_OASIS/TinyUNet__2D/__demo/default/kfold_5_fold_0', 'train_data_dir': '/Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo', 'version_dir': '/Users/zcr545/Desktop/Projects/repos/yucca_data/models/Task001_OASIS/TinyUNet__2D/__demo/default/kfold_5_fold_0/version_0', 'version': 0, 'wandb_id': 'None'}\n",
      "INFO:root:Deep Supervision Enabled: False\n",
      "INFO:root:Loading Model: 2D TinyUNet\n"
     ]
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Composing Transforms\n"
     ]
    },
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "/Users/zcr545/miniconda3/envs/yuccaenv/lib/python3.11/site-packages/tqdm/auto.py:21: TqdmWarning: IProgress not found. Please update jupyter and ipywidgets. See https://ipywidgets.readthedocs.io/en/stable/user_install.html\n",
      "  from .autonotebook import tqdm as notebook_tqdm\n",
      "INFO:root:Using 9 workers\n",
      "INFO:root:Using dataset class: <class 'yucca.data.datasets.YuccaDataset.YuccaTrainDataset'> for train/val and <class 'yucca.data.datasets.YuccaDataset.YuccaTestDataset'> for inference\n"
     ]
    }
   ],
   "source": [
    "input_dims_config = InputDimensionsConfig(\n",
    "    batch_size=config.get(\"batch_size\"), patch_size=config.get(\"patch_size\"), num_modalities=config.get(\"num_modalitites\")\n",
    ")\n",
    "task_config = TaskConfig(\n",
    "    task=config.get(\"task\"),\n",
    "    continue_from_most_recent=True,\n",
    "    experiment=config.get(\"experiment\"),\n",
    "    manager_name=\"\",\n",
    "    model_dimensions=config.get(\"dims\"),\n",
    "    model_name=config.get(\"model_name\"),\n",
    "    patch_based_training=True,\n",
    "    planner_name=config.get(\"plans_name\"),\n",
    "    split_idx=config.get(\"split_idx\"),\n",
    "    split_method=config.get(\"split_method\"),\n",
    "    split_param=config.get(\"split_param\"),\n",
    ")\n",
    "\n",
    "path_config = get_path_config(task_config=task_config)\n",
    "\n",
    "split_config = get_split_config(method=task_config.split_method, param=task_config.split_param, path_config=path_config)\n",
    "\n",
    "callback_config = get_callback_config(\n",
    "    save_dir=path_config.save_dir,\n",
    "    version_dir=path_config.version_dir,\n",
    "    experiment=task_config.experiment,\n",
    "    version=path_config.version,\n",
    "    enable_logging=False,\n",
    ")\n",
    "\n",
    "augmenter = YuccaAugmentationComposer(\n",
    "    deep_supervision=config.get(\"deep_supervision\"),\n",
    "    patch_size=input_dims_config.patch_size,\n",
    "    is_2D=True if config.get(\"dims\") == \"2D\" else False,\n",
    "    parameter_dict=generic,\n",
    "    task_type_preset=config.get(\"task_type\"),\n",
    ")\n",
    "\n",
    "\n",
    "model_module = YuccaLightningModule(\n",
    "    config=config | task_config.lm_hparams() | path_config.lm_hparams() | callback_config.lm_hparams(),\n",
    "    deep_supervision=config.get(\"deep_supervision\"),\n",
    "    learning_rate=config.get(\"learning_rate\"),\n",
    "    loss_fn=config.get(\"loss_fn\"),\n",
    "    momentum=config.get(\"momentum\"),\n",
    ")\n",
    "\n",
    "data_module = YuccaDataModule(\n",
    "    composed_train_transforms=augmenter.train_transforms,\n",
    "    composed_val_transforms=augmenter.val_transforms,\n",
    "    input_dims_config=input_dims_config,\n",
    "    train_data_dir=path_config.train_data_dir,\n",
    "    split_idx=task_config.split_idx,\n",
    "    splits_config=split_config,\n",
    "    task_type=config.get(\"task_type\"),\n",
    ")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "GPU available: True (mps), used: False\n",
      "TPU available: False, using: 0 TPU cores\n",
      "IPU available: False, using: 0 IPUs\n",
      "HPU available: False, using: 0 HPUs\n",
      "/Users/zcr545/miniconda3/envs/yuccaenv/lib/python3.11/site-packages/lightning/pytorch/trainer/setup.py:187: GPU available but not used. You can set it by doing `Trainer(accelerator='gpu')`.\n",
      "/Users/zcr545/miniconda3/envs/yuccaenv/lib/python3.11/site-packages/lightning/pytorch/trainer/connectors/checkpoint_connector.py:186: .fit(ckpt_path=\"last\") is set, but there is no last checkpoint available. No checkpoint will be loaded. HINT: Set `ModelCheckpoint(..., save_last=True)`.\n",
      "INFO:root:Setting up data for stage: TrainerFn.FITTING\n",
      "INFO:root:Training on samples: ['/Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo/OASIS_1000', '/Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo/OASIS_1001', '/Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo/OASIS_1002', '/Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo/OASIS_1008', '/Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo/OASIS_1009', '/Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo/OASIS_1010', '/Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo/OASIS_1011', '/Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo/OASIS_1012', '/Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo/OASIS_1013', '/Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo/OASIS_1014', '/Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo/OASIS_1015', '/Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo/OASIS_1036']\n",
      "INFO:root:Validating on samples: ['/Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo/OASIS_1006', '/Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo/OASIS_1007', '/Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo/OASIS_1017']\n",
      "\n",
      "  | Name          | Type             | Params\n",
      "---------------------------------------------------\n",
      "0 | train_metrics | MetricCollection | 0     \n",
      "1 | val_metrics   | MetricCollection | 0     \n",
      "2 | model         | TinyUNet         | 7.6 K \n",
      "3 | loss_fn_train | DiceCE           | 0     \n",
      "4 | loss_fn_val   | DiceCE           | 0     \n",
      "---------------------------------------------------\n",
      "7.6 K     Trainable params\n",
      "0         Non-trainable params\n",
      "7.6 K     Total params\n",
      "0.030     Total estimated model params size (MB)\n"
     ]
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Sanity Checking: |          | 0/? [00:00<?, ?it/s]"
     ]
    },
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "/Users/zcr545/miniconda3/envs/yuccaenv/lib/python3.11/site-packages/lightning/pytorch/trainer/connectors/data_connector.py:436: Consider setting `persistent_workers=True` in 'val_dataloader' to speed up the dataloader worker initialization.\n"
     ]
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Sanity Checking DataLoader 0:   0%|          | 0/2 [00:00<?, ?it/s]torch.Size([2, 3, 32, 32]) torch.Size([2, 1, 32, 32])\n",
      "Sanity Checking DataLoader 0:  50%|█████     | 1/2 [00:00<00:00, 17.28it/s]torch.Size([2, 3, 32, 32]) torch.Size([2, 1, 32, 32])\n",
      "                                                                           "
     ]
    },
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "INFO:root:Starting training with data from: /Users/zcr545/Desktop/Projects/repos/yucca_data/preprocessed/Task001_OASIS/demo\n"
     ]
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "\r"
     ]
    },
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "/Users/zcr545/miniconda3/envs/yuccaenv/lib/python3.11/site-packages/lightning/pytorch/trainer/connectors/data_connector.py:436: Consider setting `persistent_workers=True` in 'train_dataloader' to speed up the dataloader worker initialization.\n"
     ]
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Epoch 0:   0%|          | 0/2 [00:00<?, ?it/s] torch.Size([2, 3, 32, 32]) torch.Size([2, 1, 32, 32])\n",
      "Epoch 0:  50%|█████     | 1/2 [00:03<00:03,  0.30it/s, v_num=0]torch.Size([2, 3, 32, 32]) torch.Size([2, 1, 32, 32])\n",
      "Epoch 0: 100%|██████████| 2/2 [00:03<00:00,  0.59it/s, v_num=0]torch.Size([2, 3, 32, 32]) torch.Size([2, 1, 32, 32])\n",
      "torch.Size([2, 3, 32, 32]) torch.Size([2, 1, 32, 32])\n",
      "Epoch 1:   0%|          | 0/2 [00:00<?, ?it/s, v_num=0]        torch.Size([2, 3, 32, 32]) torch.Size([2, 1, 32, 32])\n",
      "Epoch 1:  50%|█████     | 1/2 [00:35<00:35,  0.03it/s, v_num=0]torch.Size([2, 3, 32, 32]) torch.Size([2, 1, 32, 32])\n",
      "Epoch 1: 100%|██████████| 2/2 [00:35<00:00,  0.06it/s, v_num=0]torch.Size([2, 3, 32, 32]) torch.Size([2, 1, 32, 32])\n",
      "torch.Size([2, 3, 32, 32]) torch.Size([2, 1, 32, 32])\n",
      "Epoch 1: 100%|██████████| 2/2 [01:20<00:00,  0.02it/s, v_num=0]"
     ]
    },
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "`Trainer.fit` stopped: `max_epochs=2` reached.\n"
     ]
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Epoch 1: 100%|██████████| 2/2 [01:20<00:00,  0.02it/s, v_num=0]\n"
     ]
    }
   ],
   "source": [
    "trainer = L.Trainer(\n",
    "    callbacks=callback_config.callbacks,\n",
    "    default_root_dir=path_config.save_dir,\n",
    "    limit_train_batches=2,\n",
    "    limit_val_batches=2,\n",
    "    log_every_n_steps=2,\n",
    "    logger=callback_config.loggers,\n",
    "    precision=\"32\",\n",
    "    profiler=callback_config.profiler,\n",
    "    enable_progress_bar=True,\n",
    "    max_epochs=2,\n",
    "    accelerator=\"cpu\",\n",
    ")\n",
    "\n",
    "\n",
    "trainer.fit(\n",
    "    model=model_module,\n",
    "    datamodule=data_module,\n",
    "    ckpt_path=\"last\",\n",
    ")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "yuccaenv",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.7"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}



"""
Task_conversion file for a dataset with 1 modality.

3 ways of Defining Train/Test splits:
  (1) Train-Test splits are defined by the folder structures of the data.

  (2) Splits are retrieved from a predefined .txt/.json/.pkl file:

    splits_file = 'path/to/data_splits.txt'
    with open(splits_file) as f:
      data = f.read()
    splits = json.loads(data)

  (3) Splits are defined using train_test_split from sklearn.model_selection.
  Always do this with a specified random seed.

2 ways of Defining Train/Val splits:

  (1) Random Train/Val splits. In this case nothing needs to be done, as these are defined at training time.

  (2) Predefined Train/Val according to a list of filenames.
  This list of files need to be defined and populated now.
  It can be generated using one of the resources mentioned above.

  It should be saved in the yucca_preprocessing_dir:

    from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, save_pickle

    maybe_mkdir_p(join(yucca_preprocessing_dir, task_name))
    save_pickle(splits, join(yucca_preprocessing_dir, task_name, 'splits.pkl'))
"""

from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.pipeline.task_conversion.utils import generate_dataset_json
import shutil
import gzip
from yucca.paths import yucca_raw_data
from tqdm import tqdm


def convert(path: str, subdir: str = "OASIS"):
    # INPUT DATA
    path = join(path, subdir)
    file_suffix = ".nii"

    # Train/Test Splits
    labels_dir_tr = join(path, "Labels", "Train")
    images_dir_tr = join(path, "Images", "Train")
    training_samples = subfiles(labels_dir_tr, join=False, suffix=file_suffix)

    labels_dir_ts = join(path, "Labels", "Test")
    images_dir_ts = join(path, "Images", "Test")
    test_samples = subfiles(labels_dir_ts, join=False, suffix=file_suffix)

    # OUTPUT DATA
    # Target names
    task_name = "Task001_OASIS"
    prefix = "OASIS"

    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    ###Populate Target Directory###
    # This is likely also the place to apply any re-orientation, resampling and/or label correction.
    for sTr in tqdm(training_samples, desc="Train"):
        serial_number = sTr[:-4]
        image_file = open(join(images_dir_tr, sTr), "rb")
        label = open(join(labels_dir_tr, sTr), "rb")
        shutil.copyfileobj(image_file, gzip.open(f"{target_imagesTr}/{prefix}_{serial_number}_000.nii.gz", "wb"))
        shutil.copyfileobj(label, gzip.open(f"{target_labelsTr}/{prefix}_{serial_number}.nii.gz", "wb"))

    for sTs in tqdm(test_samples, desc="Test"):
        serial_number = sTs[:-4]
        image_file = open(join(images_dir_ts, sTs), "rb")
        label = open(join(labels_dir_ts, sTs), "rb")
        shutil.copyfileobj(image_file, gzip.open(f"{target_imagesTs}/{prefix}_{serial_number}_000.nii.gz", "wb"))
        shutil.copyfileobj(label, gzip.open(f"{target_labelsTs}/{prefix}_{serial_number}.nii.gz", "wb"))

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        ("T1",),
        labels={0: "background", 1: "Left Hippocampus", 2: "Right Hippocampus"},
        dataset_name=task_name,
        license="hands off!",
        dataset_description="OASIS Dataset",
        dataset_reference="",
    )



from yucca.pipeline.managers.YuccaManager import YuccaManager
from yucca.data.datasets.alternative_datasets.YuccaDataset_1modality import YuccaTrainDataset_1modality


class YuccaManager_Only1Modality(YuccaManager):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(
            *args,
            **kwargs,
        )
        self.train_dataset_class = YuccaTrainDataset_1modality

    def get_plan_config(self, *args, **kwargs):
        plan_config = super().get_plan_config(*args, **kwargs)
        plan_config.num_classes = 1
        plan_config.plans["num_modalities"] = 1
        return plan_config



import wandb
import torch
from torchmetrics import MetricCollection
from torchmetrics.segmentation import GeneralizedDiceScore
from torchmetrics.classification import MulticlassF1Score
from yucca.lightning_modules.YuccaLightningModule import YuccaLightningModule


class YuccaLightningModule_onehot_labels(YuccaLightningModule):
    def __init__(
        self,
        config,
        *args,
        **kwargs,
    ):
        super().__init__(
            config,
            *args,
            **kwargs,
        )
        # self.regions_labeled = config["regions_labeled"] currently not used, but can be used during inference to go from regions -> labels

        self.train_metrics = MetricCollection(
            {
                "train/dice": GeneralizedDiceScore(num_classes=self.num_classes),
                "train/F1": MulticlassF1Score(num_classes=self.num_classes),
            },
        )

        self.val_metrics = MetricCollection(
            {
                "val/dice": GeneralizedDiceScore(num_classes=self.num_classes),
                "val/F1": MulticlassF1Score(num_classes=self.num_classes),
            },
        )

        self.loss_fn = "SigmoidDiceBCE"

    def training_step(self, batch, batch_idx):
        inputs, target, file_path = batch["image"], batch["label"], batch["file_path"]
        output = self(inputs)
        loss = self.loss_fn_train(output, target)

        if self.deep_supervision:
            # If deep_supervision is enabled output and target will be a list of (downsampled) tensors.
            # We only need the original ground truth and its corresponding prediction which is always the first entry in each list.
            output = output[0]
            target = target[0]

        output = (torch.sigmoid(output) > 0.5).long()

        metrics = self.compute_metrics(self.train_metrics, output, target)
        self.log_dict(
            {"train/loss": loss} | metrics,
            on_step=self.step_logging,
            on_epoch=self.epoch_logging,
            prog_bar=self.progress_bar,
            logger=True,
        )

        if batch_idx == 0 and wandb.run is not None and self.log_image_this_epoch is True:
            self._log_dict_of_images_to_wandb(
                {
                    "input": inputs.detach().cpu().to(torch.float32).numpy(),
                    "target": target.detach().cpu().to(torch.float32).numpy(),
                    "output": output.detach().cpu().to(torch.float32).numpy(),
                    "file_path": file_path,
                },
                log_key="train",
                task_type=self.task_type,
            )

        return loss

    def validation_step(self, batch, batch_idx):
        inputs, target, file_path = batch["image"], batch["label"], batch["file_path"]
        output = self(inputs)
        loss = self.loss_fn_val(output, target)

        output = (torch.sigmoid(output) > 0.5).long()

        metrics = self.compute_metrics(self.val_metrics, output, target)
        self.log_dict(
            {"val/loss": loss} | metrics,
            on_step=self.step_logging,
            on_epoch=self.epoch_logging,
            prog_bar=self.progress_bar,
            logger=True,
        )

        if batch_idx == 0 and wandb.run is not None and self.log_image_this_epoch is True:
            self._log_dict_of_images_to_wandb(
                {
                    "input": inputs.detach().cpu().to(torch.float32).numpy(),
                    "target": target.detach().cpu().to(torch.float32).numpy(),
                    "output": output.detach().cpu().to(torch.float32).numpy(),
                    "file_path": file_path,
                },
                log_key="val",
                task_type=self.task_type,
            )



from yucca.pipeline.planning.YuccaPlanner import YuccaPlanner


class YuccaPlannerV2(YuccaPlanner):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = str(self.__class__.__name__)
        self.get_foreground_locations_per_label = True



import numpy as np
from typing import Optional, List, Union
from batchgenerators.utilities.file_and_folder_operations import subfiles, load_pickle


def make_plans_file(
    allow_missing_modalities: bool,
    crop_to_nonzero: bool,
    classes: list[Union[int, str]],
    norm_op: str,
    modalities: List[str],
    plans_name: str,
    dataset_properties: Optional[dict] = {},
    keep_aspect_ratio_when_using_target_size: bool = True,
    preprocessor: Optional[str] = "",
    task_type: str = "segmentation",
    target_coordinate_system: str = "RAS",
    target_spacing: list = [1.0, 1.0, 1.0],
    target_size: Optional[list] = None,
    transpose_forward: list = [0, 1, 2],
    transpose_backward: list = [0, 1, 2],
    suggested_dimensionality: str = "3D",
):
    assert task_type in ["classification", "segmentation", "self-supervised"]
    plans = {}
    plans["target_coordinate_system"] = target_coordinate_system
    plans["preprocessor"] = preprocessor
    # When True all the background between the volumes and the edges of the image is removed
    plans["crop_to_nonzero"] = crop_to_nonzero

    # Change this to have different normalization schemes for all or some modalities
    plans["normalization_scheme"] = [norm_op for _ in modalities]

    plans["num_classes"] = len(classes)
    plans["num_modalities"] = len(modalities)

    plans["transpose_forward"] = transpose_forward
    plans["transpose_backward"] = transpose_backward

    # Defaults to the median spacing of the training data.
    # Change the determine_spacing() function to use different spacings
    plans["keep_aspect_ratio_when_using_target_size"] = keep_aspect_ratio_when_using_target_size
    plans["target_size"] = target_size
    plans["target_spacing"] = target_spacing
    plans["task_type"] = task_type
    plans["dataset_properties"] = dataset_properties
    plans["plans_name"] = plans_name
    plans["suggested_dimensionality"] = suggested_dimensionality
    plans["allow_missing_modalities"] = allow_missing_modalities
    return plans


def add_stats_to_plans_post_preprocessing(plans, directory):
    pkl_files = subfiles(directory, suffix=".pkl")

    new_spacings = []
    new_sizes = []
    n_cc = []
    size_cc = []
    n_foreground_locs = []
    for pkl_file in pkl_files:
        pkl_file = load_pickle(pkl_file)
        new_spacings.append(pkl_file["new_spacing"])
        new_sizes.append(pkl_file["new_size"])
        n_cc.append(pkl_file["label_cc_n"])
        n_foreground_locs.append(len(pkl_file["foreground_locations"]))
        if np.mean(pkl_file["label_cc_sizes"]) > 0:
            size_cc.append(np.mean(pkl_file["label_cc_sizes"], dtype=int))

    mean_size = np.mean(new_sizes, 0, dtype=int).tolist()
    min_size = np.min(new_sizes, 0).tolist()
    max_size = np.max(new_sizes, 0).tolist()

    if len(size_cc) > 0:
        mean_cc = np.mean(size_cc, dtype=int).tolist()
        min_cc = np.min(size_cc).tolist()
        max_cc = np.max(size_cc).tolist()
        mean_n_cc = np.mean(n_cc, dtype=int).tolist()
        min_n_cc = np.min(n_cc, initial=-1).tolist()
        max_n_cc = np.max(n_cc, initial=-1).tolist()
    else:
        mean_cc = min_cc = max_cc = mean_n_cc = min_n_cc = max_n_cc = 0

    plans["n_foreground_locs"] = n_foreground_locs
    plans["new_sizes"] = new_sizes
    plans["new_spacings"] = new_spacings
    plans["new_mean_size"] = mean_size
    plans["new_min_size"] = min_size
    plans["new_max_size"] = max_size
    plans["mean_cc_size"] = mean_cc
    plans["max_cc_size"] = max_cc
    plans["min_cc_size"] = min_cc
    plans["mean_n_cc"] = mean_n_cc
    plans["max_n_cc"] = max_n_cc
    plans["min_n_cc"] = min_n_cc
    return plans



from lightning.pytorch.callbacks import BasePredictionWriter
from yucca.functional.utils.saving import save_prediction_from_logits
from batchgenerators.utilities.file_and_folder_operations import join


class WritePredictionFromLogits(BasePredictionWriter):
    def __init__(self, output_dir, save_softmax, write_interval):
        super().__init__(write_interval)
        self.output_dir = output_dir
        self.save_softmax = save_softmax

    def write_on_batch_end(self, _trainer, _pl_module, data_dict, _batch_indices, _batch, _batch_idx, _dataloader_idx):
        # this will create N (num processes) files in `output_dir` each containing
        # the predictions of it's respective rank
        logits, properties, case_id = (
            data_dict["logits"],
            data_dict["properties"],
            data_dict["case_id"],
        )
        save_prediction_from_logits(
            logits,
            join(self.output_dir, case_id),
            properties=properties,
            save_softmax=self.save_softmax,
        )
        del data_dict



import shutil
import gzip
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.pipeline.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data, yucca_source


def convert(path: str = yucca_source, subdir: str = "LITS"):
    """INPUT DATA - Define input path and suffixes"""
    path = join(path, subdir)
    file_suffix = ".nii"

    """ OUTPUT DATA - Define the task name and prefix """
    task_name = "Task019_LITS"
    task_prefix = "LITS"

    """ Access the input data. If images are not split into train/test, and you wish to randomly 
    split the data, uncomment and adapt the following lines to fit your local path. """

    training_batches = [join(path, "Training Batch 1"), join(path, "Training Batch 2")]

    """ Then define target paths """
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    """Populate Target Directory
    This is also the place to apply any re-orientation, resampling and/or label correction."""

    for tr_batch in training_batches:
        images_dir_tr = labels_dir_tr = tr_batch
        for sTr in subfiles(images_dir_tr, join=False):
            if "volume" not in sTr:
                continue
            case_id = sTr.split("-")[-1][: -len(file_suffix)]
            src_image_file = open(join(images_dir_tr, "volume-" + case_id + file_suffix), "rb")
            src_label_file = open(join(labels_dir_tr, "segmentation-" + case_id + file_suffix), "rb")
            dst_image_file_path = f"{target_imagesTr}/{task_prefix}_{case_id}_000.nii.gz"
            dst_label_file_path = f"{target_labelsTr}/{task_prefix}_{case_id}.nii.gz"

            shutil.copyfileobj(src_image_file, gzip.open(dst_image_file_path, "wb"))
            shutil.copyfileobj(src_label_file, gzip.open(dst_label_file_path, "wb"))

    # for sTs in subfiles(images_dir_ts, suffix=file_suffix, join=False):
    #    case_id = sTs[: -len(file_suffix)]
    #    src_image_file_path = join(images_dir_ts, case_id + file_suffix)
    #    dst_image_file_path = f"{target_imagesTs}/{task_prefix}_{case_id}_000.nii.gz"
    #    shutil.copy2(src_image_file_path, dst_image_file_path)

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        modalities=("CT",),
        labels={
            0: "background",
            1: "liver",
            2: "tumor",
        },
        dataset_name=task_name,
        license="",
        dataset_description="Liver Tumor Segmentation Challenge from Miccai 2017",
        dataset_reference="https://competitions.codalab.org/competitions/17094",
    )


if __name__ == "__main__":
    convert()



import inspect


def filter_kwargs(function, kwargs):
    sig = inspect.signature(function).parameters.keys()
    return {key: value for key, value in kwargs.items() if key in sig}


def getattr_for_kwargs(obj, kwargs):
    return {key: getattr(obj, key) for key in kwargs.keys()}



import numpy as np
from typing import Optional, Union
from yucca.functional.array_operations.matrix_ops import (
    create_zero_centered_coordinate_matrix,
    deform_coordinate_matrix,
    Rx,
    Ry,
    Rz,
    Rz2D,
)
from scipy.ndimage import map_coordinates


def spatial(
    image,
    patch_size,
    p_deform,
    p_rot,
    p_rot_per_axis,
    p_scale,
    alpha,
    sigma,
    x_rot,
    y_rot,
    z_rot,
    scale_factor,
    clip_to_input_range,
    label: Optional[np.ndarray] = None,
    skip_label: bool = False,
    do_crop: bool = True,
    random_crop: bool = True,
    order: int = 3,
    cval: Optional[Union[str, int, float]] = "min",
):
    if not do_crop:
        patch_size = image.shape[2:]
    if cval == "min":
        cval = float(image.min())
    else:
        cval = cval
    assert isinstance(cval, (int, float)), f"got {cval} of type {type(cval)}"

    coords = create_zero_centered_coordinate_matrix(patch_size)
    image_canvas = np.zeros((image.shape[0], image.shape[1], *patch_size), dtype=np.float32)

    # First we apply deformation to the coordinate matrix
    if np.random.uniform() < p_deform:
        coords = deform_coordinate_matrix(coords, alpha=alpha, sigma=sigma)

    # Then we rotate the coordinate matrix around one or more axes
    if np.random.uniform() < p_rot:
        rot_matrix = np.eye(len(patch_size))
        if len(patch_size) == 2:
            rot_matrix = np.dot(rot_matrix, Rz2D(z_rot))
        else:
            if np.random.uniform() < p_rot_per_axis:
                rot_matrix = np.dot(rot_matrix, Rx(x_rot))
            if np.random.uniform() < p_rot_per_axis:
                rot_matrix = np.dot(rot_matrix, Ry(y_rot))
            if np.random.uniform() < p_rot_per_axis:
                rot_matrix = np.dot(rot_matrix, Rz(z_rot))

        coords = np.dot(coords.reshape(len(patch_size), -1).transpose(), rot_matrix).transpose().reshape(coords.shape)

    # And finally scale it
    # Scaling effect is "inverted"
    # i.e. a scale factor of 0.9 will zoom in
    if np.random.uniform() < p_scale:
        coords *= scale_factor

    if random_crop and do_crop:
        for d in range(len(patch_size)):
            crop_center_idx = [
                np.random.randint(
                    int(patch_size[d] / 2),
                    image.shape[d + 2] - int(patch_size[d] / 2) + 1,
                )
            ]
            coords[d] += crop_center_idx
    else:
        # Reversing the zero-centering of the coordinates
        for d in range(len(patch_size)):
            coords[d] += image.shape[d + 2] / 2.0 - 0.5

    # Mapping the images to the distorted coordinates
    for b in range(image.shape[0]):
        for c in range(image.shape[1]):

            img_min = image.min()
            img_max = image.max()

            image_canvas[b, c] = map_coordinates(
                image[b, c].astype(float),
                coords,
                order=order,
                mode="constant",
                cval=cval,
            ).astype(image.dtype)

            if clip_to_input_range:
                image_canvas[b, c] = np.clip(image_canvas[b, c], a_min=img_min, a_max=img_max)

    if label is not None and not skip_label:
        label_canvas = np.zeros(
            (label.shape[0], label.shape[1], *patch_size),
            dtype=np.float32,
        )

        # Mapping the labelmentations to the distorted coordinates
        for b in range(label.shape[0]):
            for c in range(label.shape[1]):
                label_canvas[b, c] = map_coordinates(label[b, c], coords, order=0, mode="constant", cval=0.0).astype(
                    label.dtype
                )
        return image_canvas, label_canvas
    return image_canvas, None



from yucca.pipeline.preprocessing.ClassificationPreprocessor import ClassificationPreprocessor
from yucca.pipeline.preprocessing.UnsupervisedPreprocessor import UnsupervisedPreprocessor
from yucca.pipeline.preprocessing.YuccaPreprocessor import YuccaPreprocessor



from yucca.data.augmentation.transforms.YuccaTransform import YuccaTransform
from yucca.functional.transforms import motion_ghosting
import numpy as np
from typing import Tuple


class MotionGhosting(YuccaTransform):
    def __init__(
        self,
        data_key="image",
        p_per_sample: float = 1.0,
        alpha=(0.85, 0.95),
        num_reps=(2, 5),
        axes=(0, 3),
        clip_to_input_range=False,
    ):
        self.data_key = data_key
        self.p_per_sample = p_per_sample
        self.alpha = alpha
        self.num_reps = num_reps
        self.axes = axes
        self.clip_to_input_range = clip_to_input_range

    @staticmethod
    def get_params(alpha: Tuple[float], num_reps: Tuple[float], axes: Tuple[float]) -> Tuple[float]:
        alpha = np.random.uniform(*alpha)
        num_reps = np.random.randint(*num_reps)
        axis = np.random.randint(*axes)
        return alpha, num_reps, axis

    def __motionGhosting__(self, image, alpha, num_reps, axis):
        image = motion_ghosting(
            image=image, alpha=alpha, num_reps=num_reps, axis=axis, clip_to_input_range=self.clip_to_input_range
        )
        return image

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape.\
            \nShould be (b, c, x, y, z) or (b, c, x, y) and is: {data_dict[self.data_key].shape}"

        for b in range(data_dict[self.data_key].shape[0]):
            for c in range(data_dict[self.data_key][b].shape[0]):
                if np.random.uniform() < self.p_per_sample:
                    alpha, num_reps, axis = self.get_params(self.alpha, self.num_reps, self.axes)
                    data_dict[self.data_key][b, c] = self.__motionGhosting__(
                        data_dict[self.data_key][b, c], alpha, num_reps, axis
                    )
        return data_dict



import numpy as np


def transpose_case(images: list, axes):
    for i in range(len(images)):
        assert len(images[i].shape) == len(axes), (
            "image and transpose axes do not match. \n"
            f"images[i].shape == {images[i].shape} \n"
            f"transpose == {axes} \n"
            f"len(images[i].shape) == {len(images[i]).shape} \n"
            f"len(transpose) == {len(axes)} \n"
        )
        images[i] = images[i].transpose(axes)
    return images


def transpose_array(image: np.ndarray, axes):
    return image.transpose(axes)



"""
https://www.meccanismocomplesso.org/en/3d-rotations-and-euler-angles-in-python/
From: https://stackoverflow.com/questions/59738230/apply-rotation-defined-by-euler-angles-to-3d-image-in-python
"""

import numpy as np
from yucca.data.augmentation.transforms.YuccaTransform import YuccaTransform
from typing import Tuple
from yucca.functional.transforms import spatial


class Spatial(YuccaTransform):
    def __init__(
        self,
        data_key="image",
        label_key="label",
        crop=False,
        cval="min",
        order=3,
        patch_size: Tuple[int] = None,
        clip_to_input_range=True,
        random_crop=True,
        p_deform_per_sample: float = 1.0,
        deform_sigma=(20, 30),
        deform_alpha=(300, 600),
        p_rot_per_sample: float = 1.0,
        p_rot_per_axis: float = 1.0,
        x_rot_in_degrees=(0.0, 10.0),
        y_rot_in_degrees=(0.0, 10.0),
        z_rot_in_degrees=(0.0, 10.0),
        p_scale_per_sample: float = 1.0,
        scale_factor=(0.85, 1.15),
        skip_label=False,
    ):
        self.data_key = data_key
        self.label_key = label_key
        self.skip_label = skip_label
        self.order = order
        self.do_crop = crop
        self.cval = cval
        self.patch_size = patch_size
        self.random_crop = random_crop
        self.clip_to_input_range = clip_to_input_range

        self.p_deform_per_sample = p_deform_per_sample
        self.deform_sigma = deform_sigma
        self.deform_alpha = deform_alpha

        self.p_rot_per_sample = p_rot_per_sample
        self.p_rot_per_axis = p_rot_per_axis
        self.x_rot_in_degrees = x_rot_in_degrees
        self.y_rot_in_degrees = y_rot_in_degrees
        self.z_rot_in_degrees = z_rot_in_degrees

        self.p_scale_per_sample = p_scale_per_sample
        self.scale_factor = scale_factor

    @staticmethod
    def get_params(
        deform_alpha: Tuple[float],
        deform_sigma: Tuple[float],
        x_rot: Tuple[float],
        y_rot: Tuple[float],
        z_rot: Tuple[float],
        scale_factor: Tuple[float],
    ) -> Tuple[float]:
        if deform_alpha:
            deform_alpha = float(np.random.uniform(*deform_alpha))
        if deform_sigma:
            deform_sigma = float(np.random.uniform(*deform_sigma))

        if x_rot:
            x_rot = float(np.random.uniform(*x_rot)) * (np.pi / 180)
        if y_rot:
            y_rot = float(np.random.uniform(*y_rot)) * (np.pi / 180)
        if z_rot:
            z_rot = float(np.random.uniform(*z_rot)) * (np.pi / 180)

        if scale_factor:
            scale_factor = float(np.random.uniform(*scale_factor))

        return deform_alpha, deform_sigma, x_rot, y_rot, z_rot, scale_factor

    def __CropDeformRotateScale__(
        self,
        data_dict,
        patch_size,
        alpha,
        sigma,
        x_rot,
        y_rot,
        z_rot,
        scale_factor,
    ):
        image, label = spatial(
            image=data_dict[self.data_key],
            patch_size=patch_size,
            p_deform=self.p_deform_per_sample,
            p_rot=self.p_rot_per_sample,
            p_rot_per_axis=self.p_rot_per_axis,
            p_scale=self.p_scale_per_sample,
            alpha=alpha,
            sigma=sigma,
            x_rot=x_rot,
            y_rot=y_rot,
            z_rot=z_rot,
            scale_factor=scale_factor,
            clip_to_input_range=self.clip_to_input_range,
            label=data_dict.get(self.label_key),
            skip_label=self.skip_label,
            do_crop=self.do_crop,
            random_crop=self.random_crop,
            order=self.order,
            cval=self.cval,
        )
        data_dict[self.data_key] = image
        if label is not None and not self.skip_label:
            data_dict[self.label_key] = label
        return data_dict

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape.\
			\nShould be (c, x, y, z) or (c, x, y) and is: {data_dict[self.data_key].shape}"

        (
            deform_alpha,
            deform_sigma,
            x_rot_rad,
            y_rot_rad,
            z_rot_rad,
            scale_factor,
        ) = self.get_params(
            deform_alpha=self.deform_alpha,
            deform_sigma=self.deform_sigma,
            x_rot=self.x_rot_in_degrees,
            y_rot=self.y_rot_in_degrees,
            z_rot=self.z_rot_in_degrees,
            scale_factor=self.scale_factor,
        )

        data_dict = self.__CropDeformRotateScale__(
            data_dict,
            self.patch_size,
            deform_alpha,
            deform_sigma,
            x_rot_rad,
            y_rot_rad,
            z_rot_rad,
            scale_factor,
        )
        return data_dict



import numpy as np


# Stolen from Batchgenerators to avoid import error caused by deprecated modules imported in
# Batchgenerators.
def augment_gamma(
    data_sample,
    gamma_range=(0.5, 2),
    invert_image=False,
    epsilon=1e-7,
    per_channel=False,
    clip_to_input_range=False,
):
    if invert_image:
        data_sample = -data_sample

    if not per_channel:
        if np.random.random() < 0.5 and gamma_range[0] < 1:
            gamma = np.random.uniform(gamma_range[0], 1)
        else:
            gamma = np.random.uniform(max(gamma_range[0], 1), gamma_range[1])
        img_min = data_sample.min()
        img_max = data_sample.max()
        img_range = img_max - img_min
        data_sample = np.power(((data_sample - img_min) / float(img_range + epsilon)), gamma) * img_range + img_min
        if clip_to_input_range:
            data_sample = np.clip(data_sample, a_min=img_min, a_max=img_max)
    else:
        for c in range(data_sample.shape[0]):
            if np.random.random() < 0.5 and gamma_range[0] < 1:
                gamma = np.random.uniform(gamma_range[0], 1)
            else:
                gamma = np.random.uniform(max(gamma_range[0], 1), gamma_range[1])
            img_min = data_sample[c].min()
            img_max = data_sample[c].max()
            img_range = img_max - img_min
            data_sample[c] = (
                np.power(((data_sample[c] - img_min) / float(img_range + epsilon)), gamma) * float(img_range + epsilon)
                + img_min
            )
            if clip_to_input_range:
                data_sample[c] = np.clip(data_sample[c], a_min=img_min, a_max=img_max)
    if invert_image:
        data_sample = -data_sample
    return data_sample



def test_configurations(setup_preprocessed_segmentation_data):
    from yucca.pipeline.configuration.configure_callbacks import get_callback_config
    from yucca.pipeline.configuration.configure_paths import get_path_config
    from yucca.pipeline.configuration.configure_plans import get_plan_config
    from yucca.pipeline.configuration.configure_task import get_task_config
    from yucca.pipeline.configuration.configure_input_dims import get_input_dims_config
    from yucca.pipeline.configuration.split_data import get_split_config
    from yucca.pipeline.configuration.configure_checkpoint import get_checkpoint_config
    from yucca.pipeline.configuration.configure_seed import seed_everything_and_get_seed_config

    # Random test
    task_config = get_task_config(task="Task000_TEST_SEGMENTATION")
    assert task_config is not None and isinstance(task_config.continue_from_most_recent, bool)

    path_config = get_path_config(task_config=task_config)
    assert path_config is not None and isinstance(path_config.version, int)

    ckpt_config = get_checkpoint_config(
        path_config=path_config,
        continue_from_most_recent=task_config.continue_from_most_recent,
        ckpt_path=None,
        current_experiment=task_config.experiment,
    )
    assert ckpt_config.ckpt_wandb_id is None or isinstance(ckpt_config.ckpt_wandb_id, str)

    seed_config = seed_everything_and_get_seed_config(ckpt_seed=ckpt_config.ckpt_seed)
    assert isinstance(seed_config.seed, int)

    plan_config = get_plan_config(
        ckpt_plans=ckpt_config.ckpt_plans,
        plans_path=path_config.plans_path,
        stage="fit",
    )
    assert plan_config is not None and plan_config.task_type in ["classification", "segmentation", "unsupervised"]

    splits_config = get_split_config(task_config.split_method, task_config.split_param, path_config)
    assert splits_config is not None and len(splits_config.train(0)) > 0

    input_dims = get_input_dims_config(
        plan=plan_config.plans,
        model_dimensions=task_config.model_dimensions,
        num_classes=plan_config.num_classes,
        model_name=task_config.model_name,
        batch_size="tiny",
        patch_size="tiny",
    )
    assert input_dims is not None and len(input_dims.patch_size) in [2, 3]

    callback_config = get_callback_config(
        enable_logging=False,
        save_dir=path_config.save_dir,
        version_dir=path_config.version_dir,
        version=path_config.version,
    )
    assert callback_config is not None and isinstance(callback_config.loggers, list)



from yucca.pipeline.managers.YuccaManager import YuccaManager
from yucca.data.datasets.YuccaCompressedDataset import YuccaCompressedTrainDataset


class YuccaManager_Compress(YuccaManager):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(
            *args,
            **kwargs,
        )
        self.train_dataset_class = YuccaCompressedTrainDataset
